#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from misc.conll18_ud_eval import load_conllu_file, evaluate

#metrics = ["Tokens", "Sentences", "Words", "UPOS", "XPOS", "UFeats", "AllTags", "Lemmas", "UAS", "LAS", "CLAS", "MLAS", "BLEX"]
#example usage:     metrics_test = conll_eval(system,gold)
#                   test_tok_f1, test_ss_f1 = metrics_test["Tokens"].f1*100., metrics_test["Sentences"].f1*100.
def conll_eval(system_file, gold_file):    
    gold_ud = load_conllu_file(gold_file)
    system_ud = load_conllu_file(system_file)
    return evaluate(gold_ud, system_ud)
    


#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dynet as dy
import numpy as np
import random
from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder


class BRNNMT:
    def __init__(self, src_we, dst_we, input_encodings, output_encodings, config):
        self.config = config
        self.losses = []
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model)
        self.src_we = src_we
        self.dst_we = dst_we
        self.input_encodings = input_encodings
        self.output_encodings = output_encodings
        # encoder
        self.encoder_fw = []
        self.encoder_bw = []
        input_size = config.input_size
        for layer_size in self.config.encoder_layers:
            self.encoder_fw.append(orthonormal_VanillaLSTMBuilder(1, input_size, layer_size, self.model))
            self.encoder_bw.append(orthonormal_VanillaLSTMBuilder(1, input_size, layer_size, self.model))
            input_size = layer_size * 2

        # decoder
        #self.decoder = []
        #for layer_size in self.config.decoder_layers:
        self.decoder=orthonormal_VanillaLSTMBuilder(config.decoder_layers, input_size+self.config.input_size, config.decoder_size, self.model)
        input_size = config.decoder_size

        # output softmax
        self.output_softmax_w = self.model.add_parameters((len(self.output_encodings.word2int) + 1, input_size))
        self.output_softmax_b = self.model.add_parameters((len(self.output_encodings.word2int) + 1))
        self.EOS = len(self.output_encodings.word2int)
        # aux WE layer
        self.aux_layer_w = self.model.add_parameters(
            (self.config.aux_we_layer_size, self.config.decoder_size))
        self.aux_layer_b = self.model.add_parameters((self.config.aux_we_layer_size))
        # aux WE projection
        self.aux_layer_proj_w = self.model.add_parameters((self.dst_we.word_embeddings_size, self.config.aux_we_layer_size))
        self.aux_layer_proj_b = self.model.add_parameters((self.dst_we.word_embeddings_size))

        # input projection
        self.word_proj_w = self.model.add_parameters((self.config.input_size, self.src_we.word_embeddings_size))
        self.word_proj_b = self.model.add_parameters((self.config.input_size))
        self.hol_we_src = self.model.add_lookup_parameters((len(self.input_encodings.word2int), self.config.input_size))
        self.hol_we_dst = self.model.add_lookup_parameters((len(self.output_encodings.word2int), self.config.input_size))
        self.special_we = self.model.add_lookup_parameters((2, self.config.input_size))

        # attention
        self.att_w1 = self.model.add_parameters(
            (self.config.encoder_layers[-1] * 2, self.config.encoder_layers[-1] * 2))
        self.att_w2 = self.model.add_parameters((self.config.encoder_layers[-1] * 2, self.config.decoder_size))
        self.att_v = self.model.add_parameters((1, self.config.encoder_layers[-1] * 2))

    def _attend(self, input_vectors, state):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * state.h()[-1]
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        loss_val = 0
        if len(self.losses) > 0:
            loss = dy.esum(self.losses)
            loss_val = loss.value()
            loss.backward()
            self.trainer.update()
        return loss_val

    def translate(self, src):
        dy.renew_cg()
        softmax_list, aux_list = self._predict(src, runtime=True)
        UNK = self.output_encodings.word2int['<UNK>']
        dst = []
        for softmax, aux in zip(softmax_list[:-1], aux_list[:-1]):
            w_index = np.argmax(softmax.npvalue())
            if w_index != UNK:
                dst.append(self.output_encodings.hol_word_list[w_index])
            else:  # this word should be searched inside the target word embeddings based on the value of AUX
                #dst.append(self.dst_we.get_closest_word(aux.value()))
                dst.append('<UNK>')

        return dst

    def learn(self, src, dst):
        softmax_list, aux_list = self._predict(src, dst=dst,  num_predictions=len(dst) + 1, runtime=False)
        for softmax, aux, entry in zip(softmax_list, aux_list, dst):
            word = entry.word.decode('utf-8').lower()
            if word in self.output_encodings.word2int:
                w_index = self.output_encodings.word2int[word]
            else:
                w_index = self.output_encodings.word2int["<UNK>"]

            w_emb, found = self.dst_we.get_word_embeddings(entry.word.decode('utf-8'))
            self.losses.append(-dy.log(dy.pick(softmax, w_index)))
            if found:
                vec1=aux
                vec2=dy.inputVector(w_emb)
                cosine = dy.dot_product(vec1, vec2) * dy.pow(dy.l2_norm(vec1) * dy.l2_norm(vec2),
                                                                       dy.scalarInput(-1))
                self.losses.append(dy.squared_distance(cosine, dy.scalarInput(1.0)))


        self.losses.append(-dy.log(dy.pick(softmax_list[-1], self.EOS)))

    def _make_input(self, list, runtime=True):
        x_list = [self.special_we[0]]
        w_emb_zero = dy.inputVector([0] * self.src_we.word_embeddings_size)

        for entry in list:
            w_emb, found = self.src_we.get_word_embeddings(entry.word.decode('utf-8'))
            if not found:
                w_emb = w_emb_zero
            else:
                w_emb = dy.inputVector(w_emb)
            word = entry.word.decode('utf-8').lower()
            if word in self.input_encodings.word2int:
                hol_emb = self.hol_we_src[self.input_encodings.word2int[word]]
            else:
                hol_emb = self.hol_we_src[self.input_encodings.word2int["<UNK>"]]

            proj_emb = self.word_proj_w.expr(update=True) * w_emb + self.word_proj_b.expr(update=True)
            if runtime:
                x = dy.tanh(proj_emb + hol_emb)
            else:
                p1 = random.random()
                p2 = random.random()
                m1 = 1
                m2 = 1
                if p1 < self.config.input_dropout_prob:
                    m1 = 0
                if p2 < self.config.input_dropout_prob:
                    m2 = 0

                scale = 1.0
                if m1 + m2 > 0:
                    scale = float(2.0) / (m1 + m2)
                m1 = dy.scalarInput(m1)
                m2 = dy.scalarInput(m2)
                scale = dy.scalarInput(scale)
                x = dy.tanh((proj_emb * m1 + hol_emb * m2) * scale)

            x_list.append(x)

        x_list.append(self.special_we[1])
        return x_list

    def _predict(self, src, dst=None, num_predictions=-1, runtime=True):
        # input
        x_list = self._make_input(src, True)
        # encoder
        for fw, bw, dropout in zip(self.encoder_fw, self.encoder_bw, self.config.encoder_layer_dropouts):
            if runtime:
                fw.set_dropouts(0, 0)
                bw.set_dropouts(0, 0)
            else:
                fw.set_dropouts(0, dropout)
                bw.set_dropouts(0, dropout)

            fw_list = fw.initial_state().transduce(x_list)
            bw_list = list(reversed(bw.initial_state().transduce(reversed(x_list))))
            x_list = [dy.concatenate([fw_value, bw_value]) for fw_value, bw_value in zip(fw_list, bw_list)]

        # decoder

        predictions_left = num_predictions
        decoder = self.decoder.initial_state().add_input(dy.inputVector([0] * (self.config.encoder_layers[-1] * 2+self.config.input_size)))
        last_dst_we=self.special_we[0]
        softmax_output = []
        aux_output = []
        pred_index=0
        while predictions_left != 0:
            predictions_left -= 1
            input = dy.concatenate([self._attend(x_list, decoder),last_dst_we])

            decoder = decoder.add_input(input)
            softmax = dy.softmax(self.output_softmax_w.expr(update=True) * decoder.output() + self.output_softmax_b.expr(update=True))
            softmax_output.append(softmax)

            proj = dy.tanh(self.aux_layer_w.expr(update=True) * decoder.output() + self.aux_layer_b.expr(update=True))
            aux = self.aux_layer_proj_w.expr(update=True) * proj + self.aux_layer_proj_b.expr(update=True)
            aux_output.append(aux)
            if runtime:
                out_we_index=np.argmax(softmax.npvalue())
                if out_we_index == self.EOS:
                    break
                last_dst_we=self.hol_we_dst[out_we_index]
            else:
                if pred_index<len(dst):
                    last_word=dst[pred_index].word.decode('utf-8').lower()
                    last_word_index=self.output_encodings.word2int['<UNK>']
                    if last_word in self.output_encodings.word2int:
                        last_word_index=self.output_encodings.word2int[last_word]
                    last_dst_we=self.hol_we_dst[last_word_index]
                    pred_index+=1
            #failsafe
            if len(softmax_output) >= 2*len(src):
                break

        return softmax_output, aux_output

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from cube.misc.misc import fopen



class CUPTDataset:
    def __init__(self, file=None):
        if file is not None:
            sys.stdout.write("Reading " + file + "... ")
            sys.stdout.flush()
            with fopen(file, "r") as f:
                lines = f.readlines()

            self.sequences = self._make_sequences(lines)
            sys.stdout.write("found " + str(len(self.sequences)) + " sequences\n")

    def _make_sequences(self, lines):
        sequences = []
        in_sequence = False
        seq = [CUPTEntry(0, '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '*', '<ROOT>',
                         '<ROOT>')]
        for line in lines:
            line = line.replace("\n", "")
            line = line.replace("\r", "")
            if (not line.startswith("#") or in_sequence) and line != '':
                parts = line.split("\t")
                s = CUPTEntry(parts[0], parts[1], parts[2], parts[3], parts[4], parts[5], parts[6], parts[10], parts[8],
                              parts[9])
                seq.append(s)
                in_sequence = True
            elif line == "":
                in_sequence = False
                if len(seq) > 0:
                    sequences.append(seq)
                seq = [CUPTEntry(0, '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '*', '<ROOT>',
                                 '<ROOT>')]

        return sequences


class CUPTEntry:
    def __init__(self, index, word, lemma, upos, xpos, attrs, head, label, deps, space_after):
        self.index, self.is_compound_entry = self._int_try_parse(index)
        self.word = word
        self.lemma = lemma
        self.upos = upos
        self.xpos = xpos
        self.attrs = attrs
        self.head, _ = self._int_try_parse(head)
        self.label = label
        self.deps = deps
        self.space_after = space_after

    def _int_try_parse(self, value):
        try:
            return int(value), False
        except ValueError:
            return value, True



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dynet as dy
import re


class CharacterNetwork:
    def __init__(self, character_embeddings_size, encodings, rnn_size=100, rnn_layers=1, embeddings_size=100,
                 model=None, runtime=False):
        if model is None:
            self.model = dy.Model()
        else:
            self.model = model

        self.encodings = encodings

        self.character_embeddings_size = character_embeddings_size
        self.embeddings_size = embeddings_size
        self.num_characters = len(encodings.char2int)
        self.character_lookup = self.model.add_lookup_parameters((self.num_characters, character_embeddings_size))

        self.rnn_fw = []
        self.rnn_bw = []
        self.rnn_layers = rnn_layers
        self.rnn_size = rnn_size
        input_size = character_embeddings_size + 3
        for _ in range(rnn_layers):
            if runtime:
                self.rnn_fw.append(dy.VanillaLSTMBuilder(1, input_size, rnn_size, self.model))
                self.rnn_bw.append(dy.VanillaLSTMBuilder(1, input_size, rnn_size, self.model))
            else:
                from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder
                self.rnn_fw.append(orthonormal_VanillaLSTMBuilder(1, input_size, rnn_size, self.model))
                self.rnn_bw.append(orthonormal_VanillaLSTMBuilder(1, input_size, rnn_size, self.model))

            input_size = rnn_size * 2
        self.linearW = self.model.add_parameters(
            (embeddings_size, rnn_size * 4))  # last state and attention over the other states
        self.linearB = self.model.add_parameters((embeddings_size))

        self.att_w1 = self.model.add_parameters((rnn_size, rnn_size * 2))
        self.att_w2 = self.model.add_parameters((rnn_size, rnn_size * 2))
        self.att_v = self.model.add_parameters((1, rnn_size))

    def compute_embeddings(self, word, runtime=True):
        x_list = []
        import sys
        import copy
        if sys.version_info[0] == 2:
            if not isinstance(word, unicode):

                uniword = unicode(word, 'utf-8')
            else:
                uniword = copy.deepcopy(word)
        else:
            uniword = copy.deepcopy(word)
        # print (uniword)
        uniword = re.sub('\d', '0', uniword)
        for i in range(len(uniword)):
            char = uniword[i]
            if char.lower() == char and char.upper() == char:
                style_emb = dy.inputVector([1.0, 0.0, 0.0])  # does not support uppercase
            elif char.lower() == char:
                style_emb = dy.inputVector([0.0, 1.0, 0.0])  # is lowercased
            else:
                style_emb = dy.inputVector([0.0, 0.0, 1.0])  # is uppercased

            char = char.lower()
            if char in self.encodings.char2int:
                x_list.append(dy.concatenate([self.character_lookup[self.encodings.char2int[char]], style_emb]))
            else:
                x_list.append(dy.concatenate([self.character_lookup[self.encodings.char2int['<UNK>']], style_emb]))

        rnn_outputs = x_list
        rnn_states_fw = None
        rnn_states_bw = None
        for rnn_fw, rnn_bw in zip(self.rnn_fw, self.rnn_bw):
            fw = []
            bw = []
            if runtime:
                rnn_fw.set_dropouts(0, 0)
                rnn_bw.set_dropouts(0, 0)
            else:
                rnn_fw.set_dropouts(0, 0.33)
                rnn_bw.set_dropouts(0, 0.33)

            rnn_fw = rnn_fw.initial_state()
            rnn_bw = rnn_bw.initial_state()
            rnn_states_fw = []
            rnn_states_bw = []
            for x in rnn_outputs:
                rnn_fw = rnn_fw.add_input(x)
                rnn_states_fw.append(rnn_fw)
                fw.append(rnn_states_fw[-1].output())
            for x in reversed(rnn_outputs):
                rnn_bw = rnn_bw.add_input(x)
                rnn_states_bw.append(rnn_bw)
                bw.append(rnn_states_bw[-1].output())
            rnn_outputs = []
            for x1, x2 in zip(fw, reversed(bw)):
                rnn_outputs.append(dy.concatenate([x1, x2]))

        attention = self._attend(rnn_outputs, rnn_states_fw[-1], rnn_states_bw[-1])

        pre_linear = dy.concatenate([fw[-1], bw[-1], attention])
        embedding = dy.tanh(self.linearW.expr(update=True) * pre_linear + self.linearB.expr(update=True))

        return embedding, rnn_outputs

    def _attend(self, input_vectors, state_fw, state_bw):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * dy.concatenate([state_fw.s()[-1], state_bw.s()[-1]])
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from cube.misc.misc import fopen
from scipy import spatial


class WordEmbeddings:
    def __init__(self, verbose=True):
        self.word2vec = {}
        self.word2ofs = {}
        self.word_embeddings_size = 0
        self.num_embeddings = 0
        self.cache_only = False
        self.file_pointer = None
        self.verbose = verbose

    def read_from_file(self, word_embeddings_file, word_list, full_load=False):
        self.word2vec = {}
        self.num_embeddings = 0
        if word_list is None and not full_load:
            self.cache_only = True
        f = fopen(word_embeddings_file, "r")
        first_line = True
        while True:
            ofs = f.tell()
            line = f.readline()
            if line == '':
                break
                # print ofs
            line = line.replace("\n", "").replace("\r", "")
            if first_line:
                first_line = False
            else:
                self.num_embeddings += 1
                if self.verbose:
                    if self.num_embeddings % 10000 == 0:
                        sys.stdout.write(
                            "  Scanned " + str(self.num_embeddings) + " word embeddings and added " + str(
                                len(self.word2vec)) + "  \n")
                parts = line.split(" ")
                if sys.version_info[0] == 2:
                    word = parts[0].decode('utf-8')
                else:
                    word = parts[0]
                if self.cache_only:
                    self.word2ofs[word] = ofs
                elif full_load or word in word_list:
                    embeddings = [float(0)] * (len(parts) - 2)

                    for zz in range(len(parts) - 2):
                        embeddings[zz] = float(parts[zz + 1])
                    self.word2vec[word] = embeddings
                self.word_embeddings_size = len(parts) - 2
        f.close()            
        if self.cache_only:            
            self.file_pointer = fopen(word_embeddings_file, "r")
            
            
    def get_word_embeddings(self, word):
        word = word.lower()
        if self.cache_only:
            if word in self.word2ofs:
                self.file_pointer.seek(self.word2ofs[word])
                line = self.file_pointer.readline()
                parts = line.split(" ")
                embeddings = [float(0)] * (len(parts) - 2)
                for zz in range(len(parts) - 2):
                    embeddings[zz] = float(parts[zz + 1])
                return embeddings, True
            else:
                return None, False

        elif word in self.word2vec:
            return self.word2vec[word], True
        else:
            return None, False

    def get_closest_word(self, vector):
        best_distance = -1.0
        best_word = '<UNK>'
        for word in self.word2vec:
            similarity = 1.0 - spatial.distance.cosine(vector, self.word2vec[word])
            if similarity > best_distance:
                best_distance = similarity
                best_word = word
        return best_word



#
# Author: Ruxandra Burtica
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

import optparse
import os
import logging

from utils import download_and_extract_archive, download_file

# Append parent dir to sys path.
parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
os.sys.path.insert(0, parent_dir)

TREEBANK_CONLL = 'https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-2515/ud-treebanks-v2.1.tgz?sequence=4&isAllowed=y'
TREEBANK_LOCATION = 'corpus/ud_treebanks'

TEST_DATA_CONLL = 'https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-2184/ud-test-v2.0-conll2017.tgz?sequence=3&isAllowed=y'
TEST_DATA_LOCATION = 'corpus/ud_test'

UD_EVAL_SCRIPT = 'https://raw.githubusercontent.com/ufal/conll2017/master/evaluation_script/conll17_ud_eval.py'
UD_EVAL_LOCATION = 'cube/misc/conll17_ud_eval.py'

EMBEDDINGS_NAME = 'wiki.{}.vec'
FACEBOOK_EMBEDDINGS_URL = 'https://s3-us-west-1.amazonaws.com/fasttext-vectors/'
FACEBOOK_EMBEDDINGS_LOCATION = 'corpus/'

logger = logging.getLogger(__name__)


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option('--language', action='store', dest='language')
    (params, _) = parser.parse_args(sys.argv)
    if not params.language:
        print(
            '\nRun the script in the following manner:\n'
            'python scripts/download_data.py --language ro\n')
        sys.exit(1)

    # Download Treebank CONLL Universal Dependencies data.
    download_and_extract_archive(TREEBANK_CONLL, TREEBANK_LOCATION)

    # Download test CONLL Universal Dependencies data.
    download_and_extract_archive(TEST_DATA_CONLL, TEST_DATA_LOCATION)

    # Download conll17_ud_eval script
    download_file(UD_EVAL_SCRIPT, UD_EVAL_LOCATION)

    # Download Facebook embeddings for the provided language.
    name = EMBEDDINGS_NAME.format(params.language)
    language_url = FACEBOOK_EMBEDDINGS_URL + name
    location = FACEBOOK_EMBEDDINGS_LOCATION + name
    download_file(language_url, location)



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dynet as dy

class CNN:
    def __init__(self, model):
        self.model = model
        self.layers = []


    def add_layer_conv(self, kernel_x, kernel_y, stride_x, stride_y, num_filters, same=True):

        if len(self.layers)==0:
            num_input_chans=1
        else:
            num_input_chans=self.layers[-1].num_output_chans

        self.layers.append(CNNConvLayer(self.model, kernel_x, kernel_y, stride_x, stride_y, num_input_chans, num_filters,same==False))

    def add_layer_pooling(self, kernel_x, kernel_y, stride_x, stride_y):
        if len(self.layers)==0:
            num_input_chans=1
        else:
            num_input_chans=self.layers[-1].num_output_chans

        self.layers.append(CNNPoolingLayer(self.model, kernel_x, kernel_y, stride_x, stride_y, num_input_chans))

    def apply(self, input_x):
        #shape=input_x.npvalue().shape
        #input_x=dy.reshape(input_x, ((shape[0], shape[1], 1))
        #print ""
        for layer in self.layers:
            #print input_x.npvalue().shape
            input_x=layer.apply(input_x)
        return input_x

class CNNConvLayer:
    def __init__(self, model, x, y, s_x, s_y, num_input_chans, num_filters, is_valid):
        self.s_x = s_x
        self.s_y = s_y
        self.is_valid=is_valid

        self.model = model
        self.kernel = self.model.add_parameters((x, y, num_input_chans, num_filters))
        #self.kernel_s = self.model.add_parameters((x, y, num_input_chans, num_filters))
        #self.kernel_t = self.model.add_parameters((x, y, num_input_chans, num_filters))
        #self.bias_s=self.model.add_parameters((num_filters))
        #self.bias_t=self.model.add_parameters((num_filters))
        self.num_output_chans = num_filters# WTF - credeam ca se mareste numarul de feature maps !!!!!#num_input_chans * num_filters

    def apply(self, x_input):
        #print "\tapplying",self.kernel.expr().npvalue().shape,"convolution"
        #output_s = dy.conv2d_bias(x_input, self.kernel_s.expr(), self.bias_s.expr(), (self.s_x, self.s_y), is_valid=self.is_valid)
        #output_t = dy.conv2d_bias(x_input, self.kernel_t.expr(), self.bias_t.expr(), (self.s_x, self.s_y), is_valid=self.is_valid)
        #return dy.cmult(dy.tanh(output_t),dy.logistic(output_s))
        output=dy.conv2d(x_input, self.kernel.expr(update=True), (self.s_x, self.s_y), is_valid=self.is_valid)
        return dy.rectify(output)

class CNNPoolingLayer:
    def __init__(self, model, x, y, s_x, s_y, num_input_chans):
        self.s_x = s_x
        self.s_y = s_y
        self.x = x
        self.y = y
        self.model = model
        self.num_output_chans = num_input_chans

    def apply(self, x_input):
        return dy.maxpooling2d(x_input, (self.x, self.y), (self.s_x, self.s_y))



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import random
import dynet as dy
from cube.generic_networks.character_embeddings import CharacterNetwork
from cube.graph.decoders import GreedyDecoder
from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder
import copy
import sys


class BDRNNParser:
    def __init__(self, parser_config, encodings, embeddings, aux_softmax_weight=0.2, runtime=False):
        self.config = parser_config
        self.encodings = encodings
        self.embeddings = embeddings
        self.decoder = GreedyDecoder()

        self.model = dy.Model()

        # self.trainer = dy.SimpleSGDTrainer(self.model)
        self.trainer = dy.AdamTrainer(self.model, alpha=2e-3, beta_1=0.9, beta_2=0.9)

        self.trainer.set_sparse_updates(False)
        self.character_network = CharacterNetwork(100, encodings, rnn_size=200, rnn_layers=1,
                                                  embeddings_size=self.config.input_embeddings_size,
                                                  model=self.model, runtime=runtime)

        self.holistic_embeddings = self.model.add_lookup_parameters(
            (len(self.encodings.word2int), self.config.input_embeddings_size))

        self.input_proj_w_word = self.model.add_parameters(
            (self.config.input_embeddings_size, self.embeddings.word_embeddings_size))
        self.input_proj_b_word = self.model.add_parameters((self.config.input_embeddings_size))

        self.unknown_word_embedding = self.model.add_lookup_parameters(
            (3, self.config.input_embeddings_size))  # for padding lexical
        self.pad_tag_embedding = self.model.add_lookup_parameters(
            (3, self.config.input_embeddings_size))  # for padding morphology

        self.bdrnn_fw = []
        self.bdrnn_bw = []

        rnn_input_size = 0
        if self.config.use_lexical:
            rnn_input_size += self.config.input_embeddings_size

        if self.config.use_morphology:
            rnn_input_size += self.config.input_embeddings_size
            self.upos_lookup = self.model.add_lookup_parameters(
                (len(self.encodings.upos2int), self.config.input_embeddings_size))
            self.xpos_lookup = self.model.add_lookup_parameters(
                (len(self.encodings.xpos2int), self.config.input_embeddings_size))
            self.attrs_lookup = self.model.add_lookup_parameters(
                (len(self.encodings.attrs2int), self.config.input_embeddings_size))

        index = 0
        aux_proj_input_size = 0
        for layer_size in self.config.layers:
            if runtime:
                self.bdrnn_fw.append(dy.VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
                self.bdrnn_bw.append(dy.VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
            else:
                self.bdrnn_fw.append(orthonormal_VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
                self.bdrnn_bw.append(orthonormal_VanillaLSTMBuilder(1, rnn_input_size, layer_size, self.model))
            rnn_input_size = layer_size * 2
            index += 1
            if index == self.config.aux_softmax_layer:
                aux_proj_input_size = rnn_input_size

        proj_input_size = self.config.layers[-1] * 2

        self.proj_arc_w_head = self.model.add_parameters((self.config.arc_proj_size, proj_input_size))
        self.proj_arc_b_head = self.model.add_parameters((self.config.arc_proj_size))
        self.proj_arc_w_dep = self.model.add_parameters((self.config.arc_proj_size, proj_input_size))
        self.proj_arc_b_dep = self.model.add_parameters((self.config.arc_proj_size))
        self.proj_label_w_head = self.model.add_parameters((self.config.label_proj_size, proj_input_size))
        self.proj_label_b_head = self.model.add_parameters((self.config.label_proj_size))
        self.proj_label_w_dep = self.model.add_parameters((self.config.label_proj_size, proj_input_size))
        self.proj_label_b_dep = self.model.add_parameters((self.config.label_proj_size))
        if not self.config.predict_morphology:
            self.aux_proj_arc_w_head = self.model.add_parameters((self.config.arc_proj_size, aux_proj_input_size))
            self.aux_proj_arc_b_head = self.model.add_parameters((self.config.arc_proj_size))
            self.aux_proj_arc_w_dep = self.model.add_parameters((self.config.arc_proj_size, aux_proj_input_size))
            self.aux_proj_arc_b_dep = self.model.add_parameters((self.config.arc_proj_size))
        else:
            self.upos_proj_w = self.model.add_parameters((self.config.label_proj_size, aux_proj_input_size))
            self.xpos_proj_w = self.model.add_parameters((self.config.label_proj_size, aux_proj_input_size))
            self.attrs_proj_w = self.model.add_parameters((self.config.label_proj_size, aux_proj_input_size))
            self.upos_proj_b = self.model.add_parameters((self.config.label_proj_size))
            self.xpos_proj_b = self.model.add_parameters((self.config.label_proj_size))
            self.attrs_proj_b = self.model.add_parameters((self.config.label_proj_size))

        self.link_b = self.model.add_parameters((1, self.config.arc_proj_size))
        self.link_w = self.model.add_parameters((self.config.arc_proj_size, self.config.arc_proj_size))

        self.label_ww = self.model.add_parameters((1, len(self.encodings.label2int)))
        self.label_w = self.model.add_parameters((len(self.encodings.label2int), self.config.label_proj_size * 2))
        self.label_bb = self.model.add_parameters((len(self.encodings.label2int)))

        if not self.config.predict_morphology:
            self.aux_link_w = self.model.add_parameters((self.config.arc_proj_size, self.config.arc_proj_size))
            self.aux_link_b = self.model.add_parameters((1, self.config.arc_proj_size))
        else:
            self.upos_softmax_w = self.model.add_parameters((len(self.encodings.upos2int), self.config.label_proj_size))
            self.xpos_softmax_w = self.model.add_parameters((len(self.encodings.xpos2int), self.config.label_proj_size))
            self.attrs_softmax_w = self.model.add_parameters(
                (len(self.encodings.attrs2int), self.config.label_proj_size))

            self.upos_softmax_b = self.model.add_parameters((len(self.encodings.upos2int)))
            self.xpos_softmax_b = self.model.add_parameters((len(self.encodings.xpos2int)))
            self.attrs_softmax_b = self.model.add_parameters((len(self.encodings.attrs2int)))
            self.lemma_softmax_b = self.model.add_parameters((len(self.encodings.char2int) + 1))
            self.lemma_softmax_casing_b = self.model.add_parameters((2))

        self.aux_softmax_weight = aux_softmax_weight
        self.batch_loss = []

    def start_batch(self):
        dy.renew_cg()
        self.batch_loss = []

    def end_batch(self):
        if len(self.batch_loss) > 0:
            loss = dy.esum(self.batch_loss)
            loss_val = loss.value()
            loss.backward()
            self.trainer.update()
            return loss_val
        else:
            return 0

    def learn(self, seq):
        # remove compound words
        tmp = []
        for ss in seq:
            if not ss.is_compound_entry:
                tmp.append(ss)
        seq = tmp
        arc_matrix, aux_arc_matrix, proj_labels, softmax_morphology = self._predict_arc(seq, runtime=False)
        gold_heads = [entry.head for entry in seq]
        gold_labels = [entry.label for entry in seq]

        softmax_labels = self._predict_label(gold_heads, proj_labels, runtime=False)

        losses = []

        for gold_head, gold_label, arc_probs, softmax_label, entry in zip(gold_heads, gold_labels,
                                                                          arc_matrix[1:],
                                                                          softmax_labels, seq):
            label_index = self.encodings.label2int[gold_label]
            losses.append(-dy.log(arc_probs[gold_head]))
            losses.append(-dy.log(dy.pick(softmax_label, label_index)))

        if not self.config.predict_morphology:
            for gold_head, aux_probs, entry in zip(gold_heads, aux_arc_matrix[
                                                               1:], seq):
                losses.append(-dy.log(aux_probs[gold_head]) * self.aux_softmax_weight)

        else:
            for softmax_morph, entry in zip(softmax_morphology, seq):
                loss_upos = -dy.log(dy.pick(softmax_morph[0], self.encodings.upos2int[entry.upos]))
                losses.append(loss_upos * (self.aux_softmax_weight / 3))

                if len(
                        self.encodings.xpos2int) > 1:  # stability check (some languages are missing attributes or XPOS, resulting in numerical overflow during backpropagation
                    loss_xpos = -dy.log(dy.pick(softmax_morph[1], self.encodings.xpos2int[entry.xpos]))
                    losses.append(loss_xpos * (self.aux_softmax_weight / 3))

                if len(
                        self.encodings.attrs2int) > 1:  # stability check (some languages are missing attributes or XPOS, resulting in numerical overflow during backpropagation
                    loss_attrs = -dy.log(dy.pick(softmax_morph[2], self.encodings.attrs2int[entry.attrs]))
                    losses.append(loss_attrs * (self.aux_softmax_weight / 3))

        loss = dy.esum(losses)
        self.batch_loss.append(loss)

    def _attend(self, input_vectors, state, aux_embeddings):
        w1 = self.lemma_att_w1.expr(update=True)
        w2 = self.lemma_att_w2.expr(update=True)
        v = self.lemma_att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * dy.concatenate([state.h()[-1], aux_embeddings])
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def tag(self, seq):
        tmp = []
        for ss in seq:
            if not ss.is_compound_entry:
                tmp.append(ss)

        # if len(tmp)<2:
        #     print "ERRRORR"
        #     for entry in seq:
        #         print str(entry.index)+"\t"+str(entry.word)
        seq = tmp

        dy.renew_cg()
        arc_matrix, aux_arc_matrix, proj_labels, softmax_morphology = self._predict_arc(seq)
        pred_heads = self.decoder.decode(arc_matrix)
        softmax_labels = self._predict_label(pred_heads, proj_labels)

        tag_list = []
        for pred_head, softmax_label in zip(pred_heads, softmax_labels):
            label_index = np.argmax(softmax_label.npvalue())
            tag = ParserTag(pred_head, self.encodings.labels[label_index], None, None, None)
            tag_list.append(tag)

        if self.config.predict_morphology:
            for tag, softmax_morph in zip(tag_list, softmax_morphology):
                tag.upos = self.encodings.upos_list[np.argmax(softmax_morph[0].npvalue())]
                tag.xpos = self.encodings.xpos_list[np.argmax(softmax_morph[1].npvalue())]
                tag.attrs = self.encodings.attrs_list[np.argmax(softmax_morph[2].npvalue())]

        return tag_list

    def _predict_label(self, heads, proj_labels, runtime=True):
        s_labels = []
        for iDep, iHead in zip(range(1, len(heads) + 1), heads):
            modw = dy.transpose(
                dy.reshape(proj_labels[iHead][1], (self.config.label_proj_size, 1)) * self.label_ww.expr(update=True))
            term1 = modw * proj_labels[iDep][0]
            term2 = self.label_w.expr(update=True) * dy.concatenate([proj_labels[iHead][1], proj_labels[iDep][0]])
            term3 = self.label_bb.expr(update=True)
            s_labels.append(dy.softmax(term1 + term2 + term3))

        return s_labels

    def _make_input(self, seq, runtime):
        x_list = []
        encoder_states_list = [None]
        # add the root
        if not self.config.use_morphology:
            x_list.append(self.unknown_word_embedding[1])
        elif not self.config.use_lexical:
            x_list.append(self.pad_tag_embedding[1])
        else:  # both lexical and morphology are used
            x_list.append(dy.concatenate(
                [self.unknown_word_embedding[1], self.pad_tag_embedding[1]]))

        for entry in seq:
            word = entry.word

            if self.config.use_lexical:
                # prepare lexical embeddings
                char_emb, encoder_states = self.character_network.compute_embeddings(word, runtime=runtime)
                encoder_states_list.append(encoder_states)
                if sys.version_info[0] == 2:
                    word_emb, found = self.embeddings.get_word_embeddings(word.decode('utf-8'))
                else:
                    word_emb, found = self.embeddings.get_word_embeddings(word)
                if not found:
                    word_emb = self.unknown_word_embedding[0]
                else:
                    word_emb = dy.tanh(
                        self.input_proj_w_word.expr(update=True) * dy.inputVector(word_emb) + self.input_proj_b_word.expr(update=True))
                if sys.version_info[0] == 2:
                    word = word.decode('utf-8').lower()
                else:
                    word = word.lower()

                if word in self.encodings.word2int:
                    holistic_emb = self.holistic_embeddings[self.encodings.word2int[word]]
                else:
                    holistic_emb = self.holistic_embeddings[self.encodings.word2int['<UNK>']]

                # dropout lexical embeddings
                if runtime:
                    w_emb = word_emb + char_emb + holistic_emb
                else:
                    p1 = random.random()
                    p2 = random.random()
                    p3 = random.random()
                    m1 = 1
                    m2 = 1
                    m3 = 1
                    if p1 < self.config.input_dropout_prob:
                        m1 = 0
                    if p2 < self.config.input_dropout_prob:
                        m2 = 0
                    if p3 < self.config.input_dropout_prob:
                        m3 = 0

                    scale = 1.0
                    if m1 + m2 + m3 > 0:
                        scale = float(3) / (m1 + m2 + m3)
                    m1 = dy.scalarInput(m1)
                    m2 = dy.scalarInput(m2)
                    m3 = dy.scalarInput(m3)
                    scale = dy.scalarInput(scale)
                    w_emb = (word_emb * m1 + char_emb * m2 + holistic_emb * m3) * scale

            if self.config.use_morphology:
                if entry.upos in self.encodings.upos2int:
                    upos_emb = self.upos_lookup[self.encodings.upos2int[entry.upos]]
                else:
                    upos_emb = dy.inputVector([0] * self.config.input_embeddings_size)
                if entry.xpos in self.encodings.xpos2int:
                    xpos_emb = self.xpos_lookup[self.encodings.xpos2int[entry.xpos]]
                else:
                    xpos_emb = dy.inputVector([0] * self.config.input_embeddings_size)
                if entry.attrs in self.encodings.attrs2int:
                    attrs_emb = self.attrs_lookup[self.encodings.attrs2int[entry.attrs]]
                else:
                    attrs_emb = dy.inputVector([0] * self.config.input_embeddings_size)
                # overwrite all dropouts. it will later be handled by "same-mask"
                t_emb = upos_emb + xpos_emb + attrs_emb
                # w_emb = word_emb + char_emb + holistic_emb

            # compose embeddings, if necessary             
            if self.config.use_lexical and self.config.use_morphology:
                if not runtime:
                    p1 = random.random()
                    p2 = random.random()
                    m1 = 1
                    m2 = 1
                    if p1 < self.config.input_dropout_prob:
                        m1 = 0
                    if p2 < self.config.input_dropout_prob:
                        m2 = 0
                    if m1 + m2 > 0:
                        scale = float(2.0) / (m1 + m2)
                    else:
                        scale = 1.0
                    scale = dy.scalarInput(scale)
                    m1 = dy.scalarInput(m1)
                    m2 = dy.scalarInput(m2)
                    x_list.append(dy.concatenate([w_emb * m1 * scale, t_emb * m2 * scale]))
                else:
                    x_list.append(dy.concatenate([w_emb, t_emb]))
            elif self.config.use_lexical:  # just use_lexical == True
                x_list.append(w_emb)
            else:  # just use_morphology == True
                x_list.append(t_emb)

        # close sequence
        if not self.config.use_morphology:
            x_list.append(self.unknown_word_embedding[2])
        elif not self.config.use_lexical:
            x_list.append(self.pad_tag_embedding[2])
        else:
            x_list.append(
                dy.concatenate(
                    [self.unknown_word_embedding[2], self.pad_tag_embedding[2]]))

        encoder_states_list.append(None)
        return x_list, encoder_states_list

    def _predict_arc(self, seq, runtime=True):
        x_list, encoder_states_list = self._make_input(seq, runtime)

        # BDLSTM
        rnn_outputs = [x_list]
        for fw, bw, dropout in zip(self.bdrnn_fw, self.bdrnn_bw, self.config.layer_dropouts):
            if runtime:
                fw.set_dropouts(0, 0)
                bw.set_dropouts(0, 0)
            else:
                fw.set_dropouts(dropout, dropout)
                bw.set_dropouts(dropout, dropout)

            fw_list = fw.initial_state().transduce(x_list)
            bw_list = list(reversed(bw.initial_state().transduce(reversed(x_list))))
            x_list = [dy.concatenate([x_fw, x_bw]) for x_fw, x_bw in zip(fw_list, bw_list)]

            rnn_outputs.append(x_list)

        # projections
        arc_projections = [[dy.tanh(self.proj_arc_w_dep.expr(update=True) * x + self.proj_arc_b_dep.expr(update=True)),
                            dy.tanh(self.proj_arc_w_head.expr(update=True) * x + self.proj_arc_b_head.expr(update=True))] for x in
                           rnn_outputs[-1]]
        label_projections = [[dy.tanh(self.proj_label_w_dep.expr(update=True) * x + self.proj_label_b_dep.expr(update=True)),
                              dy.tanh(self.proj_label_w_head.expr(update=True) * x + self.proj_label_b_head.expr(update=True))] for x in
                             rnn_outputs[-1]]
        if not runtime:
            arc_projections = [
                [dy.dropout(x1, self.config.presoftmax_mlp_dropout), dy.dropout(x2, self.config.presoftmax_mlp_dropout)]
                for x1, x2 in arc_projections]
            label_projections = [
                [dy.dropout(x1, self.config.presoftmax_mlp_dropout), dy.dropout(x2, self.config.presoftmax_mlp_dropout)]
                for x1, x2 in label_projections]
        if not self.config.predict_morphology:
            aux_arc_projections = [[dy.tanh(self.aux_proj_arc_w_dep.expr(update=True) * x + self.aux_proj_arc_b_dep.expr(update=True)),
                                    dy.tanh(self.aux_proj_arc_w_head.expr(update=True) * x + self.aux_proj_arc_b_head.expr(update=True))]
                                   for x in rnn_outputs[self.config.aux_softmax_layer]]
            if not runtime:
                aux_arc_projections = [[dy.dropout(x1, self.config.presoftmax_mlp_dropout),
                                        dy.dropout(x2, self.config.presoftmax_mlp_dropout)] for x1, x2 in
                                       aux_arc_projections]

        else:
            drp = self.config.presoftmax_mlp_dropout
            if runtime:
                drp = 0
            upos_softmax = [dy.softmax(self.upos_softmax_w.expr(update=True) * dy.dropout(dy.tanh(
                self.upos_proj_w.expr(update=True) * x + self.upos_proj_b.expr(update=True)), drp) + self.upos_softmax_b.expr(update=True)) for x in
                            rnn_outputs[self.config.aux_softmax_layer]]
            xpos_softmax = [dy.softmax(self.xpos_softmax_w.expr(update=True) * dy.dropout(dy.tanh(
                self.xpos_proj_w.expr(update=True) * x + self.xpos_proj_b.expr(update=True)), drp) + self.xpos_softmax_b.expr(update=True)) for x in
                            rnn_outputs[self.config.aux_softmax_layer]]
            attrs_softmax = [dy.softmax(self.attrs_softmax_w.expr(update=True) * dy.dropout(dy.tanh(
                self.attrs_proj_w.expr(update=True) * x + self.attrs_proj_b.expr(update=True)), drp) + self.attrs_softmax_b.expr(update=True)) for x in
                             rnn_outputs[self.config.aux_softmax_layer]]

            morphology_softmax = [[upos, xpos, attrs] for
                                  upos, xpos, attrs in
                                  zip(upos_softmax, xpos_softmax, attrs_softmax)]

        n = len(seq) + 1
        arc_matrix = [[None] * n for _ in range(n)]
        if not self.config.predict_morphology:
            aux_arc_matrix = [[None] * n for _ in range(n)]
        for iDst in range(n):
            term_bias = self.link_b.expr(update=True) * arc_projections[iDst][1]
            term_weight = self.link_w.expr(update=True) * arc_projections[iDst][1]
            if not self.config.predict_morphology:
                aux_term_bias = self.aux_link_b.expr(update=True) * aux_arc_projections[iDst][1]
                aux_term_weight = self.aux_link_w.expr(update=True) * aux_arc_projections[iDst][1]
            for iSrc in range(n):
                if iSrc != iDst:
                    attention = dy.reshape(term_weight, (1, self.config.arc_proj_size)) * arc_projections[iSrc][
                        0] + term_bias
                    arc_matrix[iSrc][iDst] = attention
                    if not self.config.predict_morphology:
                        aux_attention = dy.reshape(aux_term_weight, (1, self.config.arc_proj_size)) * \
                                        aux_arc_projections[iSrc][0] + aux_term_bias
                        aux_arc_matrix[iSrc][iDst] = aux_attention

        # compute softmax for arcs
        a_m = [[None] * n for _ in range(n)]
        if not self.config.predict_morphology:
            aux_a_m = [[None] * n for _ in range(n)]

        for iSrc in range(n):
            s_max = []
            if not self.config.predict_morphology:
                aux_s_max = []
            for iDst in range(n):
                if iSrc != iDst:
                    s_max.append(arc_matrix[iSrc][iDst])
                    if not self.config.predict_morphology:
                        aux_s_max.append(aux_arc_matrix[iSrc][iDst])
            s_max = dy.softmax(dy.concatenate(s_max))
            if not self.config.predict_morphology:
                aux_s_max = dy.softmax(dy.concatenate(aux_s_max))
            ofs = 0
            for iDst in range(n):
                if iSrc == iDst:
                    ofs = -1
                else:
                    a_m[iSrc][iDst] = s_max[iDst + ofs]
                    if not self.config.predict_morphology:
                        aux_a_m[iSrc][iDst] = aux_s_max[iDst + ofs]
        if not self.config.predict_morphology:
            return a_m, aux_a_m, label_projections, None
        else:
            return a_m, None, label_projections, morphology_softmax[1:-1]

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)

    def parse_sequences(self, sequences):
        new_sequences = []
        for sequence in sequences:
            new_sequence = copy.deepcopy(sequence)
            predicted_tags = self.tag(new_sequence)
            iOrig, iTags = 0, 0
            while iOrig < len(new_sequence):
                while new_sequence[iOrig].is_compound_entry:
                    iOrig += 1
                new_sequence[iOrig].head = predicted_tags[iTags].head
                new_sequence[iOrig].label = predicted_tags[iTags].label
                if self.config.predict_morphology == True:
                    new_sequence[iOrig].upos = predicted_tags[iTags].upos
                    new_sequence[iOrig].xpos = predicted_tags[iTags].xpos
                    new_sequence[iOrig].attrs = predicted_tags[iTags].attrs
                iTags += 1
                iOrig += 1

            new_sequences.append(new_sequence)
        return new_sequences


class ParserTag:
    def __init__(self, head, label, upos=None, xpos=None, attrs=None, lemma=None):
        self.head = head
        self.label = label
        self.upos = upos
        self.xpos = xpos
        self.attrs = attrs
        self.lemma = lemma



# -*- coding: utf-8 -*-

import sys
import os
from io_utils.model_store import ModelMetadata, ModelStore
from datetime import datetime

    
if __name__ == "__main__":

    input_models_root_folder = "/work/nlpcube"
    output_models_root_folder = "/work/nlpcube_zip"    
    version = 1.0
    
    
    
    model_store = ModelStore() 
    local_models = [os.path.basename(os.path.normpath(dI)) for dI in os.listdir(input_models_root_folder) if os.path.isdir(os.path.join(input_models_root_folder,dI))]
    #print(local_models)
    
    model_tuples = [ #folder, language_code, embedding code
("UD_Afrikaans-AfriBooms","af","af"), 
("UD_Ancient_Greek-PROIEL","grc","grc"), 
("UD_Arabic-PADT","ar","ar"), 
("UD_Armenian-ArmTDP","hy","hy"), 
("UD_Basque-BDT","eu","eu"), 
("UD_Bulgarian-BTB","bg","bg"), 
("UD_Buryat-BDT","bxr","bxr"), 
("UD_Catalan-AnCora","ca","ca"), 
("UD_Chinese-GSD","zh","zh"), 
("UD_Croatian-SET","hr","hr"), 
("UD_Czech-PDT","cs","cs"), 
("UD_Danish-DDT","da","da"), 
("UD_Dutch-Alpino","nl","nl"), 
("UD_English-EWT","en","en"), 
("UD_Estonian-EDT","et","et"), 
("UD_Finnish-TDT","fi","fi"), 
("UD_French-GSD","fr","fr"), 
("UD_Galician-CTG","gl","gl"), 
("UD_German-GSD","de","de"), 
("UD_Gothic-PROIEL","got","got"), 
("UD_Greek-GDT","el","el"), 
("UD_Hebrew-HTB","he","he"), 
("UD_Hindi-HDTB","hi","hi"), 
("UD_Hungarian-Szeged","hu","hu"), 
("UD_Indonesian-GSD","id","id"), 
("UD_Irish-IDT","ga","ga"), 
("UD_Italian-ISDT","it","it"), 
("UD_Japanese-GSD","ja","ja"), 
("UD_Kazakh-KTB","kk","kk"), 
("UD_Korean-GSD","ko","ko"), 
("UD_Kurmanji-MG","kmr","ku"), 
("UD_Latin-ITTB","la","la"), 
("UD_Latvian-LVTB","lv","lv"), 
("UD_North_Sami-Giella","sme","se"), 
("UD_Norwegian-Bokmaal","no_bokmaal","no"), 
("UD_Norwegian-Nynorsk","no_nynorsk","nn"), 
("UD_Old_Church_Slavonic-PROIEL","cu","cu"), 
("UD_Persian-Seraji","fa","fa"), 
("UD_Polish-LFG","pl","pl"), 
("UD_Portuguese-Bosque","pt","pt"), 
("UD_Romanian-RRT","ro","ro"), 
("UD_Russian-SynTagRus","ru","ru"), 
("UD_Serbian-SET","sr","sr"), 
("UD_Slovak-SNK","sk","sk"), 
("UD_Slovenian-SSJ","sl","sl"), 
("UD_Spanish-AnCora","es","es"), 
("UD_Swedish-LinES","sv","sv"), 
("UD_Swedish-Talbanken","sv","sv"), 
("UD_Turkish-IMST","tr","tr"), 
("UD_Ukrainian-IU","uk","uk"), 
("UD_Upper_Sorbian-UFAL","hsb","hsb"), 
("UD_Urdu-UDTB","ur","ur"), 
("UD_Uyghur-UDT","ug","ug"), 
("UD_Vietnamese-VTB","vi","vi")]
    
    
    
    for local_model in local_models:
        print("\nPacking: "+local_model)
        
        # find correspondent
        found = False
        language_code = ""
        embedding_code = ""
        for element in model_tuples:            
            if local_model in element[0]:
                language_code = element[1]
                embedding_code = element[2]
                found = True
                break
        if not found:
            #raise Exception("Model not found in key-store!")
            print("Model not found in key-store!")
            continue
            
        metadata = ModelMetadata()
        
        metadata.language = local_model[:local_model.find("-")]                    
        # en, ro, no_nynorsk (exception)
        metadata.language_code = language_code
        # model version: 1.0, 2.1, etc. The value is a float to perform easy comparison between versions. Format must always be #.#
        metadata.model_version = version
        # *full* link to remote embeddings file 
        metadata.embeddings_remote_link = "https://s3-us-west-1.amazonaws.com/fasttext-vectors/wiki."+embedding_code+".vec"
        # name under which the remote file will be saved under locally
        metadata.embeddings_file_name = "wiki."+embedding_code+".vec"
        # token delimiter. Must be either space (default) or "" (for languages like Japanese, Chinese, etc.)
        if language_code in "zh ja":
            metadata.token_delimiter = "" 
        else: 
            metadata.token_delimiter = " " 
        # OPTIONAL: model build date: string
        metadata.model_build_date = str(datetime.now())
        # OPTIONAL: model build source: what corpus was it built from. Ex: UD-Romanian-RRT v2.2 
        metadata.model_build_source = local_model
        # OPTIONAL: other notes, string value
        metadata.notes = "Source: ud-treebanks-v2.2"

        metadata.info()

        input_folder = os.path.join(input_models_root_folder,local_model)
        model_store.package_model(input_folder, output_models_root_folder, metadata, should_contain_tokenizer = True, should_contain_compound_word_expander = False, should_contain_lemmatizer = True, should_contain_tagger = True, should_contain_parser = True)
        
        #break # test just one package



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from cube.misc.misc import fopen

class MTDataset:
    def __init__(self, src, dst):
        sys.stdout.write("Reading files '"+src+"' and '"+dst+"'")
        sys.stdout.flush()
        source_lines=fopen(src, "r").readlines()
        destination_lines=fopen(dst,"r").readlines()
        self.sequences=self._make_sequences(source_lines, destination_lines)
        sys.stdout.write(" found "+str(len(self.sequences))+" pairs\n")

    def _make_sequences(self, src_lines, dst_lines):
        import re
        sequences=[]
        if len(src_lines)!=len(dst_lines):
            sys.stdout.write(" WARNING: datasets do not match")
        for src, dst in zip (src_lines, dst_lines):
            tok_src=src.split(" ")
            tok_dst=dst.split(" ")
            from conll import ConllEntry
            conll_src=[]
            conll_dst=[]
            for index, tok in zip(range(len(tok_src)), tok_src):
                if tok.decode('utf-8').lower()!=tok.decode('utf-8'):
                    tok='<PROPN>'
                tok=re.sub('\d','0',tok)
                if tok!='\n':
                    conll_src.append(ConllEntry(index, tok, "", "", "", "", 0, "", "", ""))

            for index, tok in zip(range(len(tok_dst)), tok_dst):
                if tok.decode('utf-8').lower()!=tok.decode('utf-8'):
                    tok='<PROPN>'
                tok = re.sub('\d', '0', tok)
                if tok != '\n':
                    conll_dst.append(ConllEntry(index, tok, "", "", "", "", 0, "", "", ""))

            sequences.append(MTSequence(conll_src, conll_dst))
        return sequences

    def to_conll_dataset(self, type='src'):
        conllDataset = FakeCONLLDataset()
        for mt_seq in self.sequences:
            if type == 'src':
                conllDataset.sequences.append(mt_seq.src)
            else:
                conllDataset.sequences.append(mt_seq.dst)
        return conllDataset

class FakeCONLLDataset():
    def __init__(self):
        self.sequences=[]


class MTSequence:
    def __init__(self, conll_src, conll_dst):
        self.src=conll_src
        self.dst=conll_dst



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import dynet_config
import optparse
import sys
sys.path.append('')
import os
import copy
from misc.misc import fopen


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option('--train', action='store', dest='train',
                      choices=['tagger', 'parser', 'lemmatizer', 'tokenizer', 'mt', 'compound', 'ner'],
                      help='select which model to train: tagger, parser, lemmatizer, tokenizer, mt, compound, ner')
    parser.add_option('--train-file', action='store', dest='train_file', help='location of the train dataset')
    parser.add_option('--dev-file', action='store', dest='dev_file', help='location of the dev dataset')
    parser.add_option('--embeddings', action='store', dest='embeddings',
                      help='location of the pre-computed word embeddings file')
    parser.add_option('--patience', action='store', dest='itters', help='no improvement stopping condition',
                      default='20', type='int')
    parser.add_option('--store', action='store', dest='output_base', help='output base for model location')
    parser.add_option('--aux-softmax', action='store', dest='aux_softmax_weight',
                      help='weight for the auxiliarly softmax', default='0.2',
                      type='float')  # should remove, it is a parser specific config
    parser.add_option("--config", action='store', dest='config', help='configuration file to load')
    parser.add_option("--test", action='store', dest='test', choices=['tagger', 'parser', 'lemmatizer', 'tokenizer'],
                      help='select what model to test')
    parser.add_option("--model", action='store', dest='model', help='location where model is stored')
    parser.add_option("--raw-train-file", action='store', dest='raw_train_file', help='location of the raw train file')
    parser.add_option("--raw-dev-file", action='store', dest='raw_dev_file', help='location of the raw dev file')
    parser.add_option("--raw-test-file", action='store', dest='raw_test_file', help='location of the raw test file')
    parser.add_option("--test-file", action='store', dest='test_file', help='location of the test dataset')
    parser.add_option("--run", action='store', dest='run',
                      help='Pipeline components: [all, tokenizer, compound, lemmatizer, tagger, parser]')
    parser.add_option("--models", action='store', dest='models', help='Folder path to where the components are stored.')

    parser.add_option("--batch-size", action='store', dest='batch_size', default='10', type='int',
                      help='number of sentences in a single batch (default=10)')
    parser.add_option("--set-mem", action='store', dest='memory', default='2048', type='int',
                      help='preallocate memory for batch training (default 2048)')
    parser.add_option("--autobatch", action='store_true', dest='autobatch',
                      help='turn on/off dynet autobatching')
    parser.add_option("--use-gpu", action='store_true', dest='gpu',
                      help='turn on/off GPU support')
    parser.add_option("--decay", action='store', dest='decay', default=0, type='float',
                      help='set value for weight decay regularization')
    parser.add_option("--params", action='store', dest='params', type='str', help='external params')

    parser.add_option("--input-file", action='store', dest='input_file',
                      help='path to the input file (either text or conllu format)')
    parser.add_option("--output-file", action='store', dest='output_file', help='path to output file to generate')
    parser.add_option("--parser-decoder", action='store', dest='decoder', default='greedy', choices=['greedy', 'mst'],
                      help='what algorithm to use for parser decoding at runtime (greedy or mst)')  # should remove, it is a parser specific config
    parser.add_option("--model-file", action='store', dest='model_base', help='what model to use')
    parser.add_option("--mt-train-src", action='store', dest='mt_train_src',
                      help='train file for source language')
    parser.add_option("--mt-train-dst", action='store', dest='mt_train_dst',
                      help='train file for destination language')
    parser.add_option("--mt-dev-src", action='store', dest='mt_dev_src',
                      help='dev file for source language')
    parser.add_option("--mt-dev-dst", action='store', dest='mt_dev_dst',
                      help='dev file for destination language')
    parser.add_option("--mt-test-src", action='store', dest='mt_test_src',
                      help='test file for source language')
    parser.add_option("--mt-test-dst", action='store', dest='mt_test_dst',
                      help='test file for destination language')
    parser.add_option("--mt-source-embeddings", action='store', dest='mt_source_embeddings',
                      help='embeddings file for sourcelanguage')
    parser.add_option("--mt-destination-embeddings", action='store', dest='mt_destination_embeddings',
                      help='embeddings file for destination language')

    parser.add_option('--model-lemmatization', action='store', dest='model_lemmatization',
                      help='precomputed lemmatization model')
    parser.add_option('--model-tokenization', action='store', dest='model_tokenization',
                      help='precomputed tokenization model')
    parser.add_option('--model-tagging', action='store', dest='model_tagging', help='precomputed tagging model')
    parser.add_option('--model-parsing', action='store', dest='model_parsing', help='precomputed parsing model')

    (params, _) = parser.parse_args(sys.argv)

    memory = int(params.memory)
    if params.autobatch:
        autobatch = True
    else:
        autobatch = False
    dynet_config.set(mem=memory, random_seed=9, autobatch=autobatch, weight_decay=params.decay)
    if params.gpu:
        dynet_config.set_gpu()

from io_utils.conll import Dataset
from io_utils.cupt import CUPTDataset
from io_utils.mt import MTDataset
from io_utils.config import TokenizerConfig
from io_utils.config import TaggerConfig
from io_utils.config import ParserConfig
from io_utils.config import LemmatizerConfig
from io_utils.config import NMTConfig
from io_utils.config import TieredTokenizerConfig
from io_utils.config import CompoundWordConfig
from io_utils.config import GDBConfig
from io_utils.embeddings import WordEmbeddings
from io_utils.encodings import Encodings
from io_utils.trainers import TokenizerTrainer
from io_utils.trainers import TaggerTrainer
from io_utils.trainers import ParserTrainer
from io_utils.trainers import LemmatizerTrainer
from io_utils.trainers import MTTrainer
from io_utils.trainers import CompoundWordTrainer
from io_utils.trainers import NERTrainer
from generic_networks.tokenizers import BDRNNTokenizer
from generic_networks.taggers import BDRNNTagger
from generic_networks.parsers import BDRNNParser
from generic_networks.lemmatizers import FSTLemmatizer
from generic_networks.translators import BRNNMT
from generic_networks.tokenizers import TieredTokenizer
from generic_networks.token_expanders import CompoundWordExpander
from generic_networks.ner import GDBNer


def parse_test(params):
    if params.test == "parser":
        print ("Running " + params.test)
        print ("==PARAMETERS==")
        print ("EMBEDDINGS: " + params.embeddings)
        print ("MODEL FILE: " + params.model_base)
        print ("DECODER: " + params.decoder)
        print ("OUTPUT: " + params.output_file)
        print ("CONFIG FILE: " + str(params.config))
        print ("==============\n")

        testset = Dataset(params.test_file)
        encodings = Encodings()
        encodings.load(params.model_base + ".encodings")
        encodings.update_wordlist(testset)
        print ("Updated word list: " + str(len(encodings.word_list)))
        config = ParserConfig(filename=params.config)
        embeddings = WordEmbeddings()
        embeddings.read_from_file(params.embeddings, encodings.word_list)
        parser = BDRNNParser(config, encodings, embeddings)
        parser.load(params.model_base + ".bestUAS")
        if params.decoder == 'mst':
            print ("!!!!!!!!!!!!!!!!!!!!!!!!!USING MST DECODER")
            from graph.decoders import MSTDecoder
            parser.decoder = MSTDecoder()
        f = fopen(params.output_file, "w")
        last_proc = 0
        index = 0
        for seq in testset.sequences:
            index += 1
            proc = index * 100 / len(testset.sequences)
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()

            rez = parser.tag(seq)
            iSeq = 0
            iRez = 0
            while iSeq < len(seq):
                while seq[iSeq].is_compound_entry:
                    iSeq += 1
                seq[iSeq].xpos = rez[iRez].xpos
                seq[iSeq].upos = rez[iRez].upos
                seq[iSeq].attrs = rez[iRez].attrs
                seq[iSeq].head = rez[iRez].head
                seq[iSeq].label = rez[iRez].label
                seq[iSeq].lemma = rez[iRez].lemma
                iSeq += 1
                iRez += 1

            for entry in seq:
                f.write(str(entry.index) + "\t" + str(entry.word) + "\t" + str(entry.lemma) + "\t" + str(
                    entry.upos) + "\t" + str(entry.xpos) + "\t" + str(entry.attrs) + "\t" + str(
                    entry.head) + "\t" + str(entry.label) + "\t" + str(entry.deps) + "\t" + str(
                    entry.space_after) + "\n")
            f.write("\n")

        f.close()
        sys.stdout.write("\n")


def parse_train(params):
    if params.train == 'mt':
        print ("Starting training for " + params.train)
        print ("==PARAMETERS==")
        print ("SRC TRAIN FILE: " + params.mt_train_src)
        print ("SRC DEV FILE: " + params.mt_dev_src)
        print ("SRC TEST FILE: " + str(params.mt_test_src))
        print ("SRC EMBEDDINGS FILE: " + params.mt_source_embeddings)
        print ("DST TRAIN FILE: " + params.mt_train_dst)
        print ("DST DEV FILE: " + params.mt_dev_dst)
        print ("DST TEST FILE: " + str(params.mt_test_dst))
        print ("DST EMBEDDINGS FILE: " + params.mt_destination_embeddings)
        print ("STOPPING CONDITION: " + str(params.itters))
        print ("OUTPUT BASE: " + params.output_base)
        print ("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print ("CONFIG FILE: " + str(params.config))
        print ("==============\n")
        trainset = MTDataset(params.mt_train_src, params.mt_train_dst)
        devset = MTDataset(params.mt_dev_src, params.mt_dev_dst)
        if params.mt_test_src and params.mt_test_dst:
            testset = MTDataset(params.mt_test_src, params.mt_test_dst)
        else:
            testset = None

        config = NMTConfig(params.config)
        sys.stdout.write("--SOURCE--\n")
        sys.stdout.flush()
        src_enc = Encodings()
        src_enc.compute(trainset.to_conll_dataset('src'), devset.to_conll_dataset('src'), word_cutoff=5)
        sys.stdout.write("--DESTINATION--\n")
        sys.stdout.flush()
        dst_enc = Encodings()
        dst_enc.compute(trainset.to_conll_dataset('dst'), devset.to_conll_dataset('dst'), word_cutoff=5)
        sys.stdout.write("Reading source embeddings\n")
        src_we = WordEmbeddings()
        src_we.read_from_file(params.mt_source_embeddings, 'label', full_load=False)
        sys.stdout.write("Reading destination embeddings\n")
        dst_we = WordEmbeddings()
        dst_we.read_from_file(params.mt_destination_embeddings, 'label', full_load=False)
        nmt = BRNNMT(src_we, dst_we, src_enc, dst_enc, config)
        trainer = MTTrainer(nmt, src_enc, dst_enc, src_we, dst_we, params.itters, trainset, devset, testset=testset)
        trainer.start_training(params.output_base, batch_size=params.batch_size)

    if params.train == "tagger":
        print ("Starting training for " + params.train)
        print ("==PARAMETERS==")
        print ("TRAIN FILE: " + params.train_file)
        print ("DEV FILE: " + params.dev_file)
        if params.test_file is not None:
            print ("TEST FILE: " + params.test_file)
        print ("EMBEDDINGS FILE: " + params.embeddings)
        print ("STOPPING CONDITION: " + str(params.itters))
        print ("OUTPUT BASE: " + params.output_base)
        print ("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print ("CONFIG FILE: " + str(params.config))
        print ("==============\n")

        trainset = Dataset(params.train_file)
        devset = Dataset(params.dev_file)
        if params.test_file:
            testset = Dataset(params.test_file)
        else:
            testset = None
        config = TaggerConfig(params.config)
        if not config._valid:
            return

        encodings = Encodings()
        encodings.compute(trainset, devset, 'label')
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)
        embeddings = WordEmbeddings()
        embeddings.read_from_file(params.embeddings, encodings.word_list)
        tagger = BDRNNTagger(config, encodings, embeddings, aux_softmax_weight=params.aux_softmax_weight)
        trainer = TaggerTrainer(tagger, encodings, params.itters, trainset, devset, testset)
        trainer.start_training(params.output_base, batch_size=params.batch_size)

    elif params.train == "parser":
        print ("Starting training for " + params.train)
        print ("==PARAMETERS==")
        print ("TRAIN FILE: " + params.train_file)
        print ("DEV FILE: " + params.dev_file)
        if params.test_file is not None:
            print ("TEST FILE: " + params.test_file)
        print ("EMBEDDINGS FILE: " + params.embeddings)
        print ("STOPPING CONDITION: " + str(params.itters))
        print ("OUTPUT BASE: " + params.output_base)
        print ("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print ("CONFIG FILE: " + str(params.config))
        print ("==============\n")

        trainset = Dataset(params.train_file)
        devset = Dataset(params.dev_file)
        if params.test_file:
            testset = Dataset(params.test_file)
        else:
            testset = None
        config = ParserConfig(params.config)
        if not config._valid:
            return
        # PARAM INJECTION 
        if params.params != None:
            parts = params.params.split(":")
            for param in parts:
                variable = param.split("=")[0]
                value = param[len(variable) + 1:]
                print("External param injection: " + variable + "=" + value)
                exec ("config.__dict__[\"" + variable + "\"] = " + value)
                # END INJECTION
        encodings = Encodings()
        encodings.compute(trainset, devset, 'label')
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)
        embeddings = WordEmbeddings()
        embeddings.read_from_file(params.embeddings, encodings.word_list)
        parser = BDRNNParser(config, encodings, embeddings, aux_softmax_weight=params.aux_softmax_weight)
        trainer = ParserTrainer(parser, encodings, params.itters, trainset, devset, testset)
        trainer.start_training(params.output_base, params.batch_size)

    elif params.train == "lemmatizer":
        print("Starting training for " + params.train)
        print("==PARAMETERS==")
        print("TRAIN FILE: " + params.train_file)
        print("DEV FILE: " + params.dev_file)
        if params.test_file is not None:
            print("TEST FILE: " + params.test_file)
        print("EMBEDDINGS FILE: " + params.embeddings)
        print("STOPPING CONDITION: " + str(params.itters))
        print("OUTPUT BASE: " + params.output_base)
        print("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print("CONFIG FILE: " + str(params.config))
        print("==============\n")

        trainset = Dataset(params.train_file)
        devset = Dataset(params.dev_file)
        if params.test_file:
            testset = Dataset(params.test_file)
        else:
            testset = None
        config = LemmatizerConfig(params.config)
        encodings = Encodings()
        encodings.compute(trainset, devset, 'label')
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)

        embeddings = None
        lemmatizer = FSTLemmatizer(config, encodings, embeddings)
        # from generic_networks.lemmatizers import BDRNNLemmatizer
        # lemmatizer = BDRNNLemmatizer(config, encodings, embeddings)
        trainer = LemmatizerTrainer(lemmatizer, encodings, params.itters, trainset, devset, testset)
        trainer.start_training(params.output_base, batch_size=params.batch_size)

    elif params.train == "compound":
        print("Starting training for " + params.train)
        print("==PARAMETERS==")
        print("TRAIN FILE: " + params.train_file)
        print("DEV FILE: " + params.dev_file)
        if params.test_file is not None:
            print("TEST FILE: " + params.test_file)
        print("EMBEDDINGS FILE: " + params.embeddings)
        print("STOPPING CONDITION: " + str(params.itters))
        print("OUTPUT BASE: " + params.output_base)
        print("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print("CONFIG FILE: " + str(params.config))
        print("==============\n")

        trainset = Dataset(params.train_file)
        devset = Dataset(params.dev_file)
        if params.test_file:
            testset = Dataset(params.test_file)
        else:
            testset = None
        config = CompoundWordConfig(params.config)
        encodings = Encodings()
        encodings.compute(trainset, devset, 'label')
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)

        embeddings = None
        expander = CompoundWordExpander(config, encodings, embeddings)
        trainer = CompoundWordTrainer(expander, encodings, params.itters, trainset, devset, testset)
        trainer.start_training(params.output_base, batch_size=params.batch_size)

    elif params.train == "tokenizer":
        print("Starting training for " + params.train)
        print("==PARAMETERS==")
        print("TRAIN FILE: " + params.train_file)
        print("RAW TRAIN FILE: " + (params.raw_train_file if params.raw_train_file is not None else "n/a"))
        print("DEV FILE: " + params.dev_file)
        print("RAW DEV FILE: " + (params.raw_dev_file if params.raw_dev_file is not None else "n/a"))
        print("TEST FILE: " + (params.test_file if params.test_file is not None else "n/a"))
        print("RAW TEST FILE: " + (params.raw_test_file if params.raw_test_file is not None else "n/a"))
        print("EMBEDDINGS FILE: " + params.embeddings)
        print("STOPPING CONDITION: " + str(params.itters))
        print("OUTPUT BASE: " + params.output_base)
        print("CONFIG FILE: " + str(params.config))
        print("==============\n")
        trainset = Dataset(params.train_file)
        devset = Dataset(params.dev_file)
        if params.test_file:
            testset = Dataset(params.test_file)
        else:
            testset = None
        from generic_networks.tokenizers import TieredTokenizer
        config = TieredTokenizerConfig(params.config)
        config.raw_test_file = params.raw_test_file
        config.base = params.output_base
        config.patience = params.itters
        if not config._valid:
            return

        encodings = Encodings()
        encodings.compute(trainset, devset, 'label')
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)
        embeddings = WordEmbeddings()
        embeddings.read_from_file(params.embeddings,
                                  None)  # setting wordlist to None triggers Word Embeddings to act as cache-only and load offsets for all words
        tokenizer = TieredTokenizer(config, encodings, embeddings)
        trainer = TokenizerTrainer(tokenizer, encodings, params.itters, trainset, devset, testset,
                                   raw_train_file=params.raw_train_file, raw_dev_file=params.raw_dev_file,
                                   raw_test_file=params.raw_test_file, gold_train_file=params.train_file,
                                   gold_dev_file=params.dev_file, gold_test_file=params.test_file)
        trainer.start_training(params.output_base, batch_size=params.batch_size)
    elif params.train == "ner":
        print ("Starting training for " + params.train)
        print ("==PARAMETERS==")
        print ("TRAIN FILE: " + params.train_file)
        print ("DEV FILE: " + params.dev_file)
        if params.test_file is not None:
            print ("TEST FILE: " + params.test_file)
        print ("EMBEDDINGS FILE: " + params.embeddings)
        print ("STOPPING CONDITION: " + str(params.itters))
        print ("OUTPUT BASE: " + params.output_base)
        print ("AUX SOFTMAX WEIGHT: " + str(params.aux_softmax_weight))
        print ("CONFIG FILE: " + str(params.config))
        print ("==============\n")

        trainset = CUPTDataset(params.train_file)
        devset = CUPTDataset(params.dev_file)
        if params.test_file:
            testset = CUPTDataset(params.test_file)
        else:
            testset = None
        config = GDBConfig(params.config)
        if not config._valid:
            return
        # PARAM INJECTION
        if params.params != None:
            parts = params.params.split(":")
            for param in parts:
                variable = param.split("=")[0]
                value = param[len(variable) + 1:]
                print("External param injection: " + variable + "=" + value)
                exec ("config.__dict__[\"" + variable + "\"] = " + value)
                # END INJECTION
        encodings = Encodings()
        encodings.compute(trainset, devset, 'label', CUPT_format=True)
        # update wordlist if testset was provided
        if params.test_file:
            encodings.update_wordlist(testset)
        embeddings = WordEmbeddings()
        embeddings.read_from_file(params.embeddings, None)
        gdbner = GDBNer(config, encodings, embeddings)
        trainer = NERTrainer(gdbner, encodings, params.itters, trainset, devset, testset)
        trainer.start_training(params.output_base, params.batch_size)


def parse_run(params):
    sys.stdout.write("\nINPUT FILE: " + params.input_file)
    sys.stdout.write("\nOUTPUT FILE: " + params.output_file)
    sys.stdout.write("\nMODELS FILE: " + params.models + "\n")
    sys.stdout.flush()

    components = params.run.split(",")
    tokenize = True if "tokenizer" in components else False
    compound = True if "compound" in components else False
    lemmatize = True if "lemmatizer" in components else False
    tag = True if "tagger" in components else False
    parse = True if "parser" in components else False

    # common elements load
    sys.stdout.write("\nLoading embeddings : " + params.embeddings + " ...\n")
    embeddings = WordEmbeddings()
    embeddings.read_from_file(params.embeddings, None)

    encodings = None
    if tokenize == True:
        if not os.path.isfile(os.path.join(params.models, "tokenizer-tok.bestAcc")):
            sys.stdout.write(
                "\n\tTokenizer model not found! (" + os.path.join(params.models, "tokenizer-tok.bestAcc") + ")")
            sys.stdout.flush()
            sys.exit(1)
        sys.stdout.write("\n\tTokenization enabled.\n")
        tokenizer_encodings = Encodings(verbose=False)
        tokenizer_encodings.load(os.path.join(params.models, "tokenizer.encodings"))
    if compound == True:
        if not os.path.isfile(os.path.join(params.models, "compound.bestAcc")):
            sys.stdout.write("\n\tCompound word expander model not found!")
            sys.stdout.flush()
            sys.exit(1)
        sys.stdout.write("\n\tCompound word expander enabled.\n")
        if encodings == None:
            encodings = Encodings(verbose=False)
            encodings.load(os.path.join(params.models, "lemmatizer.encodings"))
    if lemmatize == True:
        if not os.path.isfile(os.path.join(params.models, "lemmatizer.bestACC")):
            sys.stdout.write("\n\tLemmatization model not found!")
            sys.stdout.flush()
            sys.exit(1)
        sys.stdout.write("\n\tLemmatization enabled.\n")
        if encodings == None:
            encodings = Encodings(verbose=False)
            encodings.load(os.path.join(params.models, "lemmatizer.encodings"))
    if tag == True:
        if not os.path.isfile(os.path.join(params.models, "tagger.bestOVERALL")):
            sys.stdout.write("\n\tTagger model not found!")
            sys.stdout.flush()
            sys.exit(1)
        sys.stdout.write("\n\tTagger enabled.\n")
        if encodings == None:
            encodings = Encodings(verbose=False)
            encodings.load(os.path.join(params.models, "tagger.encodings"))
    if parse == True:
        if not os.path.isfile(os.path.join(params.models, "parser.bestUAS")):
            sys.stdout.write("\n\tParser model not found!")
            sys.stdout.flush()
            sys.exit(1)
        sys.stdout.write("\n\tParser enabled.\n")
        if encodings == None:
            encodings = Encodings(verbose=False)
            encodings.load(os.path.join(params.models, "parser.encodings"))

    sequences = None
    if tokenize:
        sys.stdout.write("\nTokenizing " + params.input_file + " ... \n\t")
        sys.stdout.flush()

        from io_utils.config import TieredTokenizerConfig
        from generic_networks.tokenizers import TieredTokenizer
        config = TieredTokenizerConfig(os.path.join(params.models, "tokenizer.conf"))
        tokenizer_object = TieredTokenizer(config, tokenizer_encodings, embeddings, runtime=True)
        tokenizer_object.load(os.path.join(params.models, "tokenizer"))

        with fopen(params.input_file, 'r') as file:
            lines = file.readlines()
        # analyze use of spaces in first part of the file
        test = "";
        useSpaces = " "
        cnt = 0
        while True:
            test = test + lines[cnt]
            # print(lines[cnt])
            if cnt >= len(lines) or cnt > 5:
                break
            cnt += 1
        if float(test.count(' ')) / float(len(test)) < 0.02:
            useSpaces = ""
        # print (str(float(test.count(' '))/float(len(test))))
        i = -1
        input_string = ""
        sequences = []
        while i < len(lines) - 1:
            i += 1
            input_string = input_string + lines[i].replace("\r", "").replace("\n", "").strip() + useSpaces
            if lines[i].strip() == "" or i == len(lines) - 1:  # end of block
                if input_string.strip() != "":
                    sequences += tokenizer_object.tokenize(input_string)
                input_string = ""

        del tokenizer_object  # free memory
    else:
        ds = Dataset(params.input_file)
        sequences = ds.sequences
    sys.stdout.write(" done\n")
    sys.stdout.flush()

    if compound:
        sys.stdout.write("\nCompound word expanding " + params.input_file + " ... \n\t")
        sys.stdout.flush()
        from generic_networks.token_expanders import CompoundWordExpander
        from io_utils.config import CompoundWordConfig
        config = CompoundWordConfig(os.path.join(params.models, "compound.conf"))
        compoundwordexpander_object = CompoundWordExpander(config, encodings, embeddings, runtime=True)
        compoundwordexpander_object.load(os.path.join(params.models, "compound.bestAcc"))
        sequences = compoundwordexpander_object.expand_sequences(sequences)
        del compoundwordexpander_object  # free memory
        sys.stdout.write(" done\n")
        sys.stdout.flush()

    if parse == True:
        sys.stdout.write("\nParsing " + params.input_file + " ... \n\t")
        sys.stdout.flush()
        from io_utils.config import ParserConfig
        from generic_networks.parsers import BDRNNParser
        config = ParserConfig(os.path.join(params.models, "parser.conf"))
        parser_object = BDRNNParser(config, encodings, embeddings, runtime=True)
        parser_object.load(os.path.join(params.models, "parser.bestUAS"))
        sequences = parser_object.parse_sequences(sequences)
        del parser_object  # free memory
        sys.stdout.write(" done\n")
        sys.stdout.flush()

    if tag == True:
        sys.stdout.write("\nTagging " + params.input_file + " ... \n\t")
        sys.stdout.flush()
        from io_utils.config import TaggerConfig
        from generic_networks.taggers import BDRNNTagger
        config = TaggerConfig(os.path.join(params.models, "tagger.conf"))
        tagger_object_UPOS = BDRNNTagger(config, encodings, embeddings, runtime=True)
        tagger_object_UPOS.load(os.path.join(params.models, "tagger.bestUPOS"))
        tagger_object_XPOS = BDRNNTagger(config, encodings, embeddings, runtime=True)
        tagger_object_XPOS.load(os.path.join(params.models, "tagger.bestXPOS"))
        tagger_object_ATTRS = BDRNNTagger(config, encodings, embeddings, runtime=True)
        tagger_object_ATTRS.load(os.path.join(params.models, "tagger.bestATTRS"))

        new_sequences = []
        for sequence in sequences:
            new_sequence = copy.deepcopy(sequence)
            predicted_tags_UPOS = tagger_object_UPOS.tag(new_sequence)
            predicted_tags_XPOS = tagger_object_XPOS.tag(new_sequence)
            predicted_tags_ATTRS = tagger_object_ATTRS.tag(new_sequence)
            for entryIndex in range(len(sequence)):
                new_sequence[entryIndex].upos = predicted_tags_UPOS[entryIndex][0]
                new_sequence[entryIndex].xpos = predicted_tags_XPOS[entryIndex][1]
                new_sequence[entryIndex].attrs = predicted_tags_ATTRS[entryIndex][2]
            new_sequences.append(new_sequence)
        sequences = copy.deepcopy(new_sequences)
        del tagger_object_UPOS  # free memory
        del tagger_object_XPOS  # free memory
        del tagger_object_ATTRS  # free memory
        sys.stdout.write(" done\n")
        sys.stdout.flush()

    if lemmatize:
        sys.stdout.write("\nLemmatizing " + params.input_file + " ... \n\t")
        sys.stdout.flush()
        from generic_networks.lemmatizers import FSTLemmatizer
        from io_utils.config import LemmatizerConfig
        config = LemmatizerConfig(os.path.join(params.models, "lemmatizer.conf"))
        lemmatizer_object = FSTLemmatizer(config, encodings, embeddings, runtime=True)
        lemmatizer_object.load(os.path.join(params.models, "lemmatizer.bestACC"))
        sequences = lemmatizer_object.lemmatize_sequences(sequences)
        del lemmatizer_object  # free memory
        sys.stdout.write(" done\n")
        sys.stdout.flush()

    output_dataset = Dataset()
    output_dataset.sequences = sequences
    output_dataset.write(params.output_file)


if params.train:
    valid = True
    if params.train != 'mt':
        if not params.train_file:
            print ("--train-file is mandatory")
            valid = False
        if not params.dev_file:
            print ("--dev-file is mandatory")
            valid = False
        if not params.embeddings:
            print ("--embeddings is mandatory")
            valid = False
        if not params.output_base:
            print ("--store is mandatory")
            valid = False
    if valid:
        parse_train(params)

if params.test:
    valid = True
    if valid:
        parse_test(params)

if params.run:
    valid = True
    if "all" in params.run:
        params.run = "tokenizer,compound,parser,tagger,lemmatizer"
    components = params.run.split(",")
    if len(components) == 0:
        print ("--run needs a list of components")
        valid = False
    if not params.embeddings:
        print ("--embeddings is mandatory")
        valid = False
    if not params.models:
        print ("--models is mandatory")
        valid = False
    if not params.input_file:
        print ("--input-file is mandatory")
        valid = False
    if not params.output_file:
        print ("--output-file is mandatory")
        valid = False
    if "lemmatizer" in params.run and not "tagger" in params.run:
        print (
            "Warning: --run needs to include a tagger to be able to lemmatize words. If not provided ensure the input file is in conllu format and includes POS tags.")
        valid = True
    if valid:
        parse_run(params)



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dynet as dy
import numpy as np
import copy
import sys
from cube.misc.misc import fopen
from .character_embeddings import CharacterNetwork


class FSTLemmatizer:
    def __init__(self, config, encodings, embeddings, runtime=False):
        self.config = config
        self.encodings = encodings
        # Bug in encodings - will be removed after UD
        self.has_bug = False
        if self.encodings.char2int[' '] != 1:
            self.has_bug = True
            import sys
            sys.stdout.write("Detected encodings BUG!")
        self.embeddings = embeddings
        self.losses = []
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model, alpha=2e-3, beta_1=0.9, beta_2=0.9)
        self.character_network = CharacterNetwork(self.config.tag_embeddings_size, encodings,
                                                  rnn_size=self.config.char_rnn_size,
                                                  rnn_layers=self.config.char_rnn_layers,
                                                  embeddings_size=self.config.char_embeddings,
                                                  model=self.model, runtime=runtime)
        self.word2lemma = {}

        self.upos_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.upos2int), self.config.tag_embeddings_size))
        self.xpos_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.xpos2int), self.config.tag_embeddings_size))
        self.attrs_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.attrs2int), self.config.tag_embeddings_size))
        self.char_lookup = self.model.add_lookup_parameters((len(self.encodings.char2int), self.config.char_embeddings))
        if runtime:
            self.rnn = dy.LSTMBuilder(self.config.rnn_layers,
                                      self.config.char_rnn_size * 2 + self.config.char_embeddings + self.config.tag_embeddings_size,
                                      self.config.rnn_size,
                                      self.model)
        else:
            from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder
            self.rnn = orthonormal_VanillaLSTMBuilder(self.config.rnn_layers,
                                                      self.config.char_rnn_size * 2 + self.config.char_embeddings + self.config.tag_embeddings_size,
                                                      self.config.rnn_size,
                                                      self.model)

        # self.att_w1 = self.model.add_parameters((200, self.config.char_rnn_size * 2))
        # self.att_w2 = self.model.add_parameters((200, self.config.rnn_size + self.config.tag_embeddings_size))
        # self.att_v = self.model.add_parameters((1, 200))

        self.start_lookup = self.model.add_lookup_parameters(
            (1, self.config.char_rnn_size * 2 + self.config.char_embeddings + self.config.tag_embeddings_size))

        self.softmax_w = self.model.add_parameters((len(self.encodings.char2int) + 3, self.config.rnn_size))
        self.softmax_b = self.model.add_parameters((len(self.encodings.char2int) + 3))

        ofs = len(self.encodings.char2int)
        self.label2int = {}
        self.label2int['<EOS>'] = ofs
        self.label2int['<COPY>'] = ofs + 1
        self.label2int['<INC>'] = ofs + 2

    def _attend(self, input_vectors, state, embeddings):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * dy.concatenate([state.h()[-1], embeddings])
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def _predict(self, word, upos, xpos, attrs, max_predictions=0, runtime=True, gs_labels=None):
        char_emb, states = self.character_network.compute_embeddings(word, runtime=runtime)

        softmax_list = []
        m1, m2, m3 = 0, 0, 0
        zero_vec = dy.vecInput(self.config.tag_embeddings_size)
        if upos in self.encodings.upos2int:
            upos_emb = self.upos_lookup[self.encodings.upos2int[upos]]
            m1 = 1
        else:
            upos_emb = zero_vec

        if xpos in self.encodings.xpos2int:
            xpos_emb = self.xpos_lookup[self.encodings.xpos2int[xpos]]
            m2 = 1
        else:
            xpos_emb = zero_vec

        if attrs in self.encodings.attrs2int:
            attrs_emb = self.attrs_lookup[self.encodings.attrs2int[attrs]]
            m3 = 1
        else:
            attrs_emb = zero_vec

        scale = float(4.0) / (m1 + m2 + m3 + 1.0)

        scale = dy.scalarInput(scale)
        tag_emb = (upos_emb + xpos_emb + attrs_emb + char_emb) * scale
        rnn = self.rnn.initial_state().add_input(self.start_lookup[0])
        num_predictions = 0
        i_src = 0
        i_labels = 0
        while num_predictions < max_predictions:
            # attention = self._attend(states, rnn, tag_emb)

            input = dy.concatenate([char_emb, states[i_src], tag_emb])
            rnn = rnn.add_input(input)

            softmax = dy.softmax(self.softmax_w.expr(update=True) * rnn.output() + self.softmax_b.expr(update=True))
            softmax_list.append(softmax)
            num_predictions += 1
            if runtime:
                l_index = np.argmax(softmax.npvalue())
                if l_index == self.label2int['<EOS>']:
                    break
                elif l_index == self.label2int['<INC>'] and i_src < len(states) - 1:
                    i_src += 1
            else:
                if gs_labels[i_labels] == '<INC>' and i_src < len(states) - 1:
                    i_src += 1
            i_labels += 1

        return softmax_list

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        total_loss = 0
        if len(self.losses) > 0:
            loss = dy.esum(self.losses)
            total_loss = loss.value()
            loss.backward()
            self.trainer.update()
        self.losses = []
        return total_loss

    def learn(self, seq):
        for entry in seq:
            if entry.upos != 'NUM' and entry.upos != 'PROPN':
                # print entry.word+"\t"+entry.lemma
                import sys
                if sys.version_info[0] == 2:
                    y_real = self._compute_transduction_states(unicode(entry.word, 'utf-8').lower(),
                                                               unicode(entry.lemma, 'utf-8').lower())
                else:
                    y_real = self._compute_transduction_states(entry.word.lower(),
                                                               entry.lemma.lower())
                # print y_real
                losses = []
                n_chars = len(y_real)
                # print entry.word, entry.lemma
                # print y_real
                softmax_output_list = self._predict(entry.word, entry.upos, entry.xpos, entry.attrs,
                                                    max_predictions=n_chars, runtime=False, gs_labels=y_real)
                # print unilemma.encode('utf-8')#, softmax_output_list
                for softmax, y_target in zip(softmax_output_list, y_real):
                    if y_target in self.label2int:
                        losses.append(-dy.log(dy.pick(softmax, self.label2int[y_target])))
                    elif y_target in self.encodings.char2int:
                        losses.append(-dy.log(dy.pick(softmax, self.encodings.char2int[y_target])))

                if len(losses) > 0:
                    loss = dy.esum(losses)

                self.losses.append(loss)

    def _compute_transduction_states(self, source, destination):
        a = np.zeros((len(source) + 1, len(destination) + 1))
        for i in range(len(source) + 1):
            a[i, 0] = i

        for i in range(len(destination) + 1):
            a[0, i] = i

        for i in range(1, len(source) + 1):
            for j in range(1, len(destination) + 1):
                cost = 0
                if source[i - 1] != destination[j - 1]:
                    cost = 1
                m = min([a[i - 1, j - 1], a[i - 1, j], a[i, j - 1]])
                a[i, j] = m + cost

        alignments = [-1] * len(destination)

        i = len(source)
        j = len(destination)
        while i > 1 or j > 1:
            if source[i - 1] == destination[j - 1]:
                alignments[j - 1] = i - 1
            if i == 1:
                j -= 1
            elif j == 1:
                i -= 1
            else:
                if a[i - 1, j - 1] <= a[i - 1, j] and a[i - 1, j - 1] <= a[i, j - 1]:
                    i -= 1
                    j -= 1
                elif a[i - 1][j] <= a[i - 1, j - 1] and a[i - 1, j] <= a[i, j - 1]:
                    i -= 1
                else:
                    j -= 1
        if source[i - 1] == destination[j - 1]:
            alignments[j - 1] = i - 1

        y_pred = []
        index_src = 0
        index_dst = 0
        while index_dst < len(destination):
            if alignments[index_dst] == index_src:
                y_pred.append("<COPY>")
                index_dst += 1
            elif alignments[index_dst] == -1:
                if destination[index_dst] == "\t":
                    y_pred.append("<TOK>")
                    index_dst += 1
                else:
                    y_pred.append(destination[index_dst])
                    index_dst += 1
            else:
                y_pred.append("<INC>")
                index_src += 1

        y_pred.append("<EOS>")
        return y_pred

    def tag(self, seq):
        dy.renew_cg()
        lemmas = []
        for entry in seq:
            if entry.upos == 'NUM' or entry.upos == 'PROPN':
                import sys
                if sys.version_info[0] == 2:

                    lemma = entry.word.decode('utf-8')
                else:
                    lemma = entry.word
            else:
                # check dictionary
                import sys
                if sys.version_info[0]==2:
                    key = entry.word.decode('utf-8').lower().encode('utf-8') + "\t" + entry.lemma
                else:
                    key = entry.word.lower() + "\t" + entry.lemma                    
                if key in self.word2lemma:
                    if sys.version_info[0]==2:
                        lemma=unicode(self.word2lemma[key],'utf-8')
                    else:                        
                        lemma=copy.deepcopy(self.word2lemma[key])
                else:
                    if sys.version_info[0]==2:
                        uniword = unicode(entry.word, 'utf-8')
                    else: 
                        uniword=copy.deepcopy(entry.word)
                        

                    softmax_output_list = self._predict(uniword, entry.upos, entry.xpos, entry.attrs,
                                                        max_predictions=500, runtime=True)
                    lemma = ""
                    src_index = 0
                    for softmax in softmax_output_list[:-1]:
                        label_index = np.argmax(softmax.npvalue())
                        if label_index == self.label2int['<COPY>'] and src_index < len(uniword):
                            lemma += uniword[src_index]
                        elif label_index == self.label2int['<INC>'] or label_index == self.label2int['<EOS>']:
                            src_index += 1
                        elif label_index < len(self.encodings.characters):
                            # if self.has_bug and label_index >= self.encodings.char2int[' ']:
                            #     label_index += 1
                            lemma += self.encodings.characters[label_index]
            # print entry.word+"\t"+lemma.encode('utf-8')
            if entry.upos != 'PROPN':
                lemmas.append(lemma.lower())
            else:
                lemmas.append(lemma)
        return lemmas

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)
        dict_path = path.replace(".bestAcc", ".dict")
        import os.path
        if os.path.exists(dict_path):
            self.load_dict(dict_path)

    def load_dict(self, path):
        #print ("Loading lemma dictionary")
        with fopen(path, "r") as f:
            lines = f.readlines()
            for line in lines:
                parts = line.strip().split('\t')
                if len(parts) == 5:
                    if sys.version_info[0] == 2:
                        word = unicode(parts[0], 'utf-8').lower().encode('utf-8')
                    else:
                        word = parts[0].lower()
                    upos = parts[1]
                    key = word + '\t' + upos
                    self.word2lemma[key] = parts[4]
        #print ("Loaded " + str(len(self.word2lemma)) + " pairs")

    def lemmatize_sequences(self, sequences):
        new_sequences = []
        for sequence in sequences:
            new_sequence = copy.deepcopy(sequence)
            predicted_lemmas = self.tag(new_sequence)

            for entry, lemma in zip(new_sequence, predicted_lemmas):
                if not entry.is_compound_entry:
                    entry.lemma = lemma if lemma is not None else "_"  # lemma.encode('utf-8')
                else:
                    entry.lemma = "_"
            # for entryIndex, lemma in enumerate(predicted_lemmas):
            #    new_sequence[entryIndex].lemma = lemma if lemma is not None else "_"
            new_sequences.append(new_sequence)
        return new_sequences


class BDRNNLemmatizer:
    def __init__(self, lemmatizer_config, encodings, embeddings, runtime=False):
        self.config = lemmatizer_config
        self.encodings = encodings
        # Bug in encodings - this will be removed after UD Shared Task
        self.has_bug = False
        if self.encodings.char2int[' '] != 1:
            self.has_bug = True
        self.embeddings = embeddings
        self.losses = []

        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model, alpha=2e-3, beta_1=0.9, beta_2=0.9)

        self.character_network = CharacterNetwork(self.config.tag_embeddings_size, encodings,
                                                  rnn_size=self.config.char_rnn_size,
                                                  rnn_layers=self.config.char_rnn_layers,
                                                  embeddings_size=self.config.char_embeddings,
                                                  model=self.model, runtime=runtime)

        self.upos_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.upos2int), self.config.tag_embeddings_size))
        self.xpos_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.xpos2int), self.config.tag_embeddings_size))
        self.attrs_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.attrs2int), self.config.tag_embeddings_size))
        self.char_lookup = self.model.add_lookup_parameters((len(self.encodings.char2int), self.config.char_embeddings))

        if runtime:
            self.rnn = dy.LSTMBuilder(self.config.rnn_layers,
                                      self.config.char_rnn_size * 2 + self.config.char_embeddings, self.config.rnn_size,
                                      self.model)
        else:
            from generic_networks.utils import orthonormal_VanillaLSTMBuilder
            self.rnn = orthonormal_VanillaLSTMBuilder(self.config.rnn_layers,
                                                      self.config.char_rnn_size * 2 + self.config.char_embeddings,
                                                      self.config.rnn_size,
                                                      self.model)

        self.att_w1 = self.model.add_parameters((200, self.config.char_rnn_size * 2))
        self.att_w2 = self.model.add_parameters((200, self.config.rnn_size + self.config.tag_embeddings_size))
        self.att_v = self.model.add_parameters((1, 200))

        self.start_lookup = self.model.add_lookup_parameters(
            (1, self.config.char_rnn_size * 2 + self.config.char_embeddings))

        self.softmax_w = self.model.add_parameters((len(self.encodings.char2int) + 1, self.config.rnn_size))
        self.softmax_b = self.model.add_parameters((len(self.encodings.char2int) + 1))
        self.softmax_casing_w = self.model.add_parameters((2, self.config.rnn_size))
        self.softmax_casing_b = self.model.add_parameters((2))

    def _attend(self, input_vectors, state, embeddings):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * dy.concatenate([state.s()[-1], embeddings])
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def _predict(self, word, upos, xpos, attrs, num_chars=0, gs_chars=None):
        if num_chars == 0:
            runtime = True
        else:
            runtime = False

        char_emb, states = self.character_network.compute_embeddings(word, runtime=runtime)

        num_predictions = 0
        softmax_list = []
        m1, m2, m3 = 0, 0, 0
        zero_vec = dy.vecInput(self.config.tag_embeddings_size)
        if upos in self.encodings.upos2int:
            upos_emb = self.upos_lookup[self.encodings.upos2int[upos]]
            m1 = 1
        else:
            upos_emb = zero_vec

        if xpos in self.encodings.xpos2int:
            xpos_emb = self.xpos_lookup[self.encodings.xpos2int[xpos]]
            m2 = 1
        else:
            xpos_emb = zero_vec

        if attrs in self.encodings.attrs2int:
            attrs_emb = self.attrs_lookup[self.encodings.attrs2int[attrs]]
            m3 = 1
        else:
            attrs_emb = zero_vec

        scale = float(4.0) / (m1 + m2 + m3 + 1.0)

        scale = dy.scalarInput(scale)
        tag_emb = (upos_emb + xpos_emb + attrs_emb + char_emb) * scale
        rnn = self.rnn.initial_state().add_input(self.start_lookup[0])
        char_emb = dy.inputVector([0] * self.config.char_embeddings)

        while True:
            attention = self._attend(states, rnn, tag_emb)

            input = dy.concatenate([attention, char_emb])
            rnn = rnn.add_input(input)

            softmax = dy.softmax(self.softmax_w.expr(update=True) * rnn.output() + self.softmax_b.expr(update=True))
            softmax_casing = dy.softmax(self.softmax_casing_w.expr(update=True) * rnn.output() + self.softmax_casing_b.expr(update=True))
            softmax_list.append([softmax, softmax_casing])
            if num_chars == 0:
                s_index = np.argmax(softmax.npvalue())
                if s_index == len(self.encodings.char2int):
                    break
                char_emb = self.char_lookup[s_index]
            else:
                if num_predictions < len(gs_chars):
                    char = gs_chars[num_predictions]
                    if char in self.encodings.char2int:
                        char_emb = self.char_lookup[self.encodings.char2int[char]]
                    else:
                        char_emb = self.char_lookup[self.encodings.char2int["<UNK>"]]

            num_predictions += 1
            if num_predictions == num_chars or num_predictions > 255:
                break

        return softmax_list

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        total_loss = 0
        if len(self.losses) > 0:
            loss = dy.esum(self.losses)
            total_loss = loss.value()
            loss.backward()
            self.trainer.update()
        self.losses = []
        return total_loss

    def learn(self, seq):

        for entry in seq:
            if entry.upos != 'NUM' and entry.upos != 'PROPN':
                losses = []
                unilemma = unicode(entry.lemma, 'utf-8')
                n_chars = len(unilemma)
                softmax_output_list = self._predict(entry.word, entry.upos, entry.xpos, entry.attrs,
                                                    num_chars=n_chars + 1, gs_chars=unilemma)
                # print unilemma.encode('utf-8')#, softmax_output_list
                for softmax, char in zip(softmax_output_list[:-1], unilemma):

                    char_index = -1
                    if char.lower() == char:
                        casing = 0
                    else:
                        casing = 1
                    char = char.lower()
                    if char in self.encodings.char2int:
                        char_index = self.encodings.char2int[char]
                    if char_index != -1:
                        losses.append(-dy.log(dy.pick(softmax[0], char_index)))
                    losses.append(-dy.log(dy.pick(softmax[1], casing)))
                    # print np.argmax(softmax[0].npvalue()), char_index, softmax

                losses.append(-dy.log(dy.pick(softmax_output_list[-1][0], len(self.encodings.char2int))))
                loss = dy.esum(losses)
                self.losses.append(loss)

    def tag(self, seq):
        dy.renew_cg()
        lemmas = []
        for entry in seq:
            if entry.upos == 'NUM' or entry.upos == 'PROPN':
                if sys.version_info[0] == 2:
                    lemma = entry.word.decode('utf-8')
                else:
                    lemma = entry.word
            else:
                softmax_output_list = self._predict(entry.word, entry.upos, entry.xpos, entry.attrs)
                lemma = ""
                for softmax in softmax_output_list[:-1]:
                    char_index = np.argmax(softmax[0].npvalue())
                    if char_index < len(self.encodings.characters):
                        char = self.encodings.characters[char_index]
                        if np.argmax(softmax[1].npvalue()) == 1:
                            char = char.upper()
                        lemma += char
            lemmas.append(lemma)
        return lemmas

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)



#
# Author: Ruxandra Burtica
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import shutil
import tarfile

import requests


def download_file(url, file_name):
    # First remove the archive_name file if it exists.
    if os.path.exists(file_name):
        os.remove(file_name)
    request = requests.get(url)
    with open(file_name, "wb") as fd:
        fd.write(request.content)


def download_and_extract_archive(url, file_location):
    archive_name = file_location + '.tgz'

    # Download file
    download_file(url, archive_name)

    # Extract from archive and remove archive.
    tar = tarfile.open(archive_name)
    tar.extractall(path=file_location)
    tar.close()
    os.remove(archive_name)

    # Move content to the pre-defined location.
    directories = [
        name for name in os.listdir(file_location)
        if os.path.isdir(os.path.join(file_location, name))]
    sub_directory = directories[0]
    for filename in os.listdir(os.path.join(file_location, sub_directory)):
        shutil.move(
            os.path.join(file_location, sub_directory, filename),
            os.path.join(file_location, filename))



#
# Authors: Tiberiu Boros, Stefan Daniel Dumitrescu
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import ast
from builtins import object, super
from cube.misc.misc import fopen
import collections

if sys.version_info[0] == 2:
    import ConfigParser
else:
    import configparser


class Config(object):
    """Generic base class that implements load/save utilities."""

    def __init__(self):
        """Call to set config object name."""
        self.__config__ = self.__class__.__name__

    def _auto_cast(self, s):
        """Autocasts string s to its original type."""
        try:
            return ast.literal_eval(s)
        except:
            return s

    def save(self, filename):
        """Save configuration to file."""
        sorted_dict = collections.OrderedDict(sorted(self.__dict__.items()))  # sort dictionary
        if sys.version_info[0] == 2:
            config = ConfigParser.ConfigParser()
        else:
            config = configparser.ConfigParser()
        config.add_section(self.__config__)  # write header
        if sys.version_info[0] == 2:
            items = sorted_dict.iteritems()
        else:
            items = sorted_dict.items()
        for k, v in items:  # for python3 use .items()
            if not k.startswith("_"):  # write only non-private properties
                if isinstance(v, float):  # if we are dealing with a float
                    str_v = str(v)
                    if "e" not in str_v and "." not in str_v:  # stop possible confusion with an int by appending a ".0"
                        v = str_v + ".0"
                v = str(v)
                config.set(self.__config__, k, v)
        with fopen(filename, 'w') as cfgfile:
            config.write(cfgfile)

    def load(self, filename):
        """Load configuration from file."""
        if sys.version_info[0] == 2:
            config = ConfigParser.ConfigParser()
        else:
            config = configparser.ConfigParser()
        config.read(filename)
        # check to see if the config file has the appropriate section
        if not config.has_section(self.__config__):
            sys.stderr.write(
                "ERROR: File \"" + filename + "\" is not a valid configuration file for the selected task: Missing section [" + self.__config__ + "]!\n")
            sys.exit(1)
        for k, v in config.items(self.__config__):
            self.__dict__[k] = self._auto_cast(v)


class TokenizerConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()

        self.base = ""
        # encoder-char        
        self.char_vocabulary_size = -1  # to be calculated when first training
        self.char_embedding_size = 100
        self.char_generic_feature_vocabulary_size = 2
        self.char_generic_feature_embedding_size = 5

        self.encoder_char_input_attribute_dropout = 0.
        self.encoder_char_lstm_size = 200

        # next-chars
        self.next_chars_embedding_size = 100
        self.next_chars_window_size = 10

        # encoder-word        
        self.encoder_word_input_w2i_array = {}
        self.encoder_word_vocab_size = 0  # ref
        self.encoder_word_embedding_size = 0  # ref
        self.encoder_word_lstm_size = 200
        # decoder 
        self.decoder_attribute_dropout = 0.33
        self.decoder_hidden_size = 20

        self.dropout_rate = 0
        # extra        
        self.patience = -1
        self.tokenize_maximum_sequence_length = 500  # how much to run predict on, at a time

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + "\n")
            self.load(filename)

        self._valid = True


class TaggerConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        self.layers = [200, 200]
        self.layer_dropouts = [0.5, 0.5]
        self.aux_softmax_layer = 1
        self._valid = True
        self.input_dropout_prob = 0.33
        self.presoftmax_mlp_layers = [500]
        self.presoftmax_mlp_dropouts = [0.5]
        self.input_size = 100

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)

        if verbose:
            print ("INPUT SIZE:", self.input_size)
            print ("LAYERS:", self.layers)
            print ("LAYER DROPOUTS:", self.layer_dropouts)
            print ("AUX SOFTMAX POSITION:", self.aux_softmax_layer)
            print ("INPUT DROPOUT PROB:", self.input_dropout_prob)
            print ("PRESOFTMAX MLP LAYERS:", self.presoftmax_mlp_layers)
            print ("PRESOFTMAX MLP DROPOUT:", self.presoftmax_mlp_dropouts)

        if self.aux_softmax_layer > len(self.layers) - 1 or self.aux_softmax_layer == 0:
            print (
                "Configuration error: aux softmax layer must be placed after the first layer and before the final one.")
            self._valid = False


class ParserConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        self.layers = [300, 300, 200, 200, 200]
        self.layer_dropouts = [0.33, 0.33, 0.33, 0.33, 0.33]
        self.aux_softmax_layer = 2
        self._valid = True
        self.input_dropout_prob = 0.33
        self.arc_proj_size = 100
        self.label_proj_size = 400
        self.presoftmax_mlp_dropout = 0.33
        self.predict_morphology = True
        self.use_morphology = False
        self.use_lexical = True
        self.input_embeddings_size = 100

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)

        if verbose:
            print ("LAYERS:", self.layers)
            print ("LAYER DROPOUTS:", self.layer_dropouts)
            print ("AUX SOFTMAX POSITION:", self.aux_softmax_layer)
            print ("INPUT DROPOUT PROB:", self.input_dropout_prob)
            print ("ARC PROJECTION SIZE:", self.arc_proj_size)
            print ("LABEL PROJECTION SIZE:", self.label_proj_size)
            print ("PRESOFTMAX MLP DROPOUT:", self.presoftmax_mlp_dropout)
            print ("JOINTLY PARSE AND PREDICT MORPHOLOGY:", self.predict_morphology)
            print ("USE MORPHOLOGY AS INPUT:", self.use_morphology)
            print ("INPUT EMBEDDINGS SIZE:", self.input_embeddings_size)

        if self.aux_softmax_layer > len(self.layers) - 1 or self.aux_softmax_layer == 0:
            print (
                "Configuration error: aux softmax layer must be placed after the first layer and before the final one.")
            self._valid = False

        if self.use_morphology and self.predict_morphology:
            print ("Configuration error: you are using morphology to predict morphology.")
            self._valid = False


class LemmatizerConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        self.rnn_size = 200
        self.rnn_layers = 2
        self.char_embeddings = 100
        self.char_rnn_size = 200
        self.char_rnn_layers = 2
        self.tag_embeddings_size = 100

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)


class NMTConfig(Config):
    def __init__(self, filename=None):
        super().__init__()
        self.encoder_layers = [300, 300]
        self.encoder_layer_dropouts = [0.33, 0.33]
        self.decoder_layers = 2
        self.decoder_size = 300
        self.decoder_dropout = 0.33
        self.input_size = 100
        self.aux_we_layer_size = 100
        self.input_dropout_prob = 0.33

        if filename is None:
            sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)


class TieredTokenizerConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        # sentece splitting
        self.ss_char_embeddings_size = 100
        self.ss_char_peek_count = 5
        self.ss_mlp_layers = [100]
        self.ss_mlp_dropouts = [0.33]
        self.ss_lstm_size = 64
        self.ss_lstm_layers = 1
        self.ss_lstm_dropout = 0.33
        self.ss_peek_lstm_size = 64
        self.ss_peek_lstm_layers = 1
        self.ss_peek_lstm_dropout = 0.33
        # tokenization
        self.tok_char_embeddings_size = 100
        self.tok_word_embeddings_size = 100
        self.tok_mlp_layers = [100]
        self.tok_mlp_dropouts = [0.33]
        self.tok_char_lstm_layers = 2
        self.tok_char_lstm_size = 200
        self.tok_char_lstm_dropout = 0.33
        self.tok_word_lstm_layers = 2
        self.tok_word_lstm_size = 200
        self.tok_word_lstm_dropout = 0.33
        self.tok_char_peek_lstm_layers = 2
        self.tok_char_peek_lstm_size = 200
        self.tok_char_peek_lstm_dropout = 0.33

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)

        self._valid = True


class CompoundWordConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        self.character_embeddings_size = 100
        self.encoder_size = 200
        self.encoder_layers = 2
        self.decoder_size = 200
        self.decoder_layers = 2

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)


class GDBConfig(Config):
    def __init__(self, filename=None, verbose=False):
        super().__init__()
        self.use_char_embeddings = True
        self.char_rnn_layers = 2
        self.char_rnn_size = 100
        self.embeddings_size = 100
        self.arc_rnn_layers = [200, 200]
        self.label_rnn_size = 100
        self.proj_size = 100

        if filename is None:
            if verbose:
                sys.stdout.write("No configuration file supplied. Using default values.\n")
        else:
            if verbose:
                sys.stdout.write("Reading configuration file " + filename + " \n")
            self.load(filename)

        self._valid = True



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dynet as dy
import numpy as np


def get_link(seq, iSrc, iDst):
    l1 = seq[iSrc].label
    l2 = seq[iDst].label

    if iSrc == 0 and l2 != '*':
        return 1
    if iDst == 0 and l1 != '*':
        return 1

    if l1 == "*" or l2 == "*":
        return 0

    pp_l1 = l1.split(';')
    pp_l2 = l2.split(';')
    for l1 in pp_l1:
        for l2 in pp_l2:
            ppl1 = l1.split(':')[0]
            ppl2 = l2.split(':')[0]
            if ppl1 == ppl2:
                return 1
    return 0


def _has_index(index, label):
    parts = label.split(";")
    for part in parts:
        pp = part.split(":")
        if pp[0] == str(index):
            return True
    return False


class GDBNer:
    def __init__(self, config, encodings, embeddings, runtime=False):
        self.model = dy.Model()
        self.config = config
        self.encodings = encodings
        self.embeddings = embeddings

        self.trainer = dy.AdamTrainer(self.model, alpha=2e-3, beta_1=0.9, beta_2=0.9)

        self.word_lookup = self.model.add_lookup_parameters((len(self.encodings.word2int), self.config.embeddings_size))
        self.upos_lookup = self.model.add_lookup_parameters((len(self.encodings.upos2int), self.config.embeddings_size))
        self.xpos_lookup = self.model.add_lookup_parameters((len(self.encodings.xpos2int), self.config.embeddings_size))
        self.attrs_lookup = self.model.add_lookup_parameters(
            (len(self.encodings.attrs2int), self.config.embeddings_size))

        if self.config.use_char_embeddings:
            from cube.generic_networks.character_embeddings import CharacterNetwork
            self.character_network = CharacterNetwork(self.config.embeddings_size, encodings,
                                                      rnn_size=self.config.char_rnn_size,
                                                      rnn_layers=self.config.char_rnn_layers,
                                                      embeddings_size=self.config.embeddings_size, model=self.model,
                                                      runtime=runtime)

        self.we_proj = self.model.add_parameters(
            (self.config.embeddings_size, self.embeddings.word_embeddings_size))

        self.encoder_fw = []
        self.encoder_bw = []

        lstm_builder = dy.VanillaLSTMBuilder
        if not runtime:
            from cube.generic_networks.utils import orthonormal_VanillaLSTMBuilder
            lstm_builder = orthonormal_VanillaLSTMBuilder

        input_size = self.config.embeddings_size

        for layer_size in self.config.arc_rnn_layers:
            self.encoder_fw.append(lstm_builder(1, input_size, layer_size, self.model))
            self.encoder_bw.append(lstm_builder(1, input_size, layer_size, self.model))
            input_size = layer_size * 2

        self.link_w = self.model.add_parameters((1, self.config.proj_size * 2))
        self.link_b = self.model.add_parameters((1))

        self.label_decoder = lstm_builder(1, self.config.proj_size, self.config.label_rnn_size, self.model)

        self.proj_w1 = self.model.add_parameters((self.config.proj_size, input_size))
        self.proj_w2 = self.model.add_parameters((self.config.proj_size, input_size))
        self.proj_w3 = self.model.add_parameters((self.config.proj_size, input_size))
        self.proj_b1 = self.model.add_parameters((self.config.proj_size))
        self.proj_b2 = self.model.add_parameters((self.config.proj_size))
        self.proj_b3 = self.model.add_parameters((self.config.proj_size))

        self.label_w = self.model.add_parameters((len(self.encodings.label2int), self.config.label_rnn_size))
        self.label_b = self.model.add_parameters((len(self.encodings.label2int)))

        self.losses = []

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        total_loss = 0
        if len(self.losses) > 0:
            loss = dy.esum(self.losses)
            total_loss = loss.value()
            loss.backward()
            self.trainer.update()
        return total_loss

    def _make_input(self, seq, runtime=False):
        seq_input = []
        zero_vec_tag = dy.inputVector([0 for i in range(self.config.embeddings_size)])
        for entry in seq:
            word = entry.word
            upos = entry.upos
            xpos = entry.xpos
            attrs = entry.attrs

            word_vec, found = self.embeddings.get_word_embeddings(word)
            if not found:
                word_vec, found = self.embeddings.get_word_embeddings('</s>')

            word_vector = self.we_proj.expr(update=True) * dy.inputVector(word_vec)

            tag_mult = 1.0
            if upos in self.encodings.upos2int:
                upos_vec = self.upos_lookup[self.encodings.upos2int[upos]]
            else:
                upos_vec = zero_vec_tag
                tag_mult += 1.0

            if xpos in self.encodings.xpos2int:
                xpos_vec = self.xpos_lookup[self.encodings.xpos2int[xpos]]
            else:
                xpos_vec = zero_vec_tag
                tag_mult += 1.0

            if attrs is self.encodings.attrs2int:
                attrs_vec = self.attrs_lookup[self.encodings.attrs2int[attrs]]
            else:
                attrs_vec = zero_vec_tag
                tag_mult += 1.0

            tag_vector = (upos_vec + xpos_vec + attrs_vec) * dy.scalarInput(tag_mult)

            if self.config.use_char_embeddings:
                char_vector, states = self.character_network.compute_embeddings(word, runtime=runtime)
            else:
                char_vector = zero_vec_tag

            p1 = np.random.random()
            p2 = np.random.random()
            p3 = np.random.random()
            if not self.config.use_char_embeddings:
                p3 = 1.0

            scale = 1.0
            if not runtime:
                if p1 < 0.34:
                    word_vector = zero_vec_tag
                    scale += 1.0
                if p2 < 0.34:
                    tag_vector = zero_vec_tag
                    scale += 1.0
                if p3 < 0.34:
                    char_vector = zero_vec_tag
                    scale += 1.0
            seq_input.append((word_vector + tag_vector + char_vector) * dy.scalarInput(scale))

        return seq_input

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model.populate(path)

    def tag(self, seq):
        dy.renew_cg()
        from io_utils.cupt import CUPTEntry
        seq = [CUPTEntry(0, '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '<ROOT>', '*', '<ROOT>',
                         '<ROOT>')] + seq  # append root
        output, proj_x = self._predict(seq, runtime=True)
        return self._decode(output, proj_x)

    def _predict(self, seq, runtime=True):
        x_list = self._make_input(seq, runtime=runtime)

        for fw, bw in zip(self.encoder_fw, self.encoder_bw):
            x_fw = fw.initial_state().transduce(x_list)
            x_bw = list(reversed(bw.initial_state().transduce(reversed(x_list))))
            x_list = [dy.concatenate([x1, x2]) for x1, x2 in zip(x_fw, x_bw)]

        proj_x1 = [dy.tanh(self.proj_w1.expr(update=True) * x + self.proj_b1.expr(update=True)) for x in x_list]
        proj_x2 = [dy.tanh(self.proj_w2.expr(update=True) * x + self.proj_b2.expr(update=True)) for x in x_list]
        proj_x3 = [dy.tanh(self.proj_w3.expr(update=True) * x + self.proj_b3.expr(update=True)) for x in x_list]

        output = []
        for iSrc in range(len(seq)):
            out_row = []
            for iDst in range(len(seq)):
                if iDst > iSrc:
                    x = dy.concatenate([proj_x1[iSrc], proj_x2[iDst]])
                    out_row.append(dy.logistic(self.link_w.expr(update=True) * x + self.link_b.expr(update=True)))
                else:
                    out_row.append(None)
            output.append(out_row)
        return output, proj_x3

    def _get_gs_chains(self, seq):
        indices = []
        for row in seq:
            if row.label != "*":
                parts = row.label.split(";")
                for part in parts:
                    pp = part.split(":")
                    expr_index = int(pp[0])
                    if expr_index not in indices:
                        indices.append(expr_index)

        chains = []
        labels = []
        for index in indices:
            first = True
            lst = []
            label = ""
            i = 0
            for row in seq:
                if _has_index(index, row.label):
                    if first:
                        first = False
                        parts = row.label.split(";")
                        for part in parts:
                            pp = part.split(":")
                            if len(pp) == 1:
                                print(str(row.index) + "\t" + row.word + "\t" + row.label)
                            if pp[0] == str(index):
                                label = pp[1]
                                break

                    lst.append(i)
                i += 1
            if label == "":
                for row in seq:
                    print (row.orig_line)
            chains.append(lst)
            labels.append(label)

        return chains, labels

    def _valid(self, a, current_nodes, node):
        for other_node in current_nodes:
            if a[node, other_node] == 0:
                return False
        return True

    def _backtrack(self, a, current_nodes, solutions):
        recursed = False
        for i in range(a.shape[0]):
            if a[current_nodes[-1], i] == 1:
                if i not in current_nodes:
                    if self._valid(a, current_nodes, i):
                        current_nodes.append(i)  # push
                        recursed = True
                        self._backtrack(a, current_nodes, solutions)
                        current_nodes = current_nodes[:-1]  # pop
        if not recursed and len(current_nodes) > 1:
            import copy
            solutions.append(copy.deepcopy(current_nodes))

    def learn(self, seq):
        output, proj_x3 = self._predict(seq, runtime=False)

        # arcs
        for iSrc in range(len(seq)):
            for iDst in range(len(seq)):
                if iDst > iSrc:
                    o = output[iSrc][iDst]  # the softmax portion
                    t = get_link(seq, iSrc, iDst)
                    # if t==1:
                    # self.losses.append(-dy.log(dy.pick(o, t)))
                    self.losses.append(dy.binary_log_loss(o, dy.scalarInput(t)))

        # labels
        gs_chains, labels = self._get_gs_chains(seq)

        for chain, label in zip(gs_chains, labels):
            label_rnn = self.label_decoder.initial_state()
            for index in chain:
                label_rnn = label_rnn.add_input(proj_x3[index])
            label_softmax = dy.softmax(
                self.label_w.expr(update=True) * label_rnn.output() + self.label_b.expr(update=True))
            self.losses.append(-dy.log(dy.pick(label_softmax, self.encodings.label2int[label])))

    def _decode(self, output, proj_x):
        expressions = []
        labels = []
        a = np.zeros((len(output), len(output)))
        for iSrc in range(len(output)):
            for iDst in range(len(output)):
                if iDst > iSrc:
                    if output[iSrc][iDst].value() >= 0.5:
                        a[iSrc][iDst] = 1
                        a[iDst][iSrc] = 1

        for iSrc in range(len(output)):
            exprs = []
            current_nodes = [iSrc]
            self._backtrack(a, current_nodes, exprs)
            [expr.sort() for expr in exprs]
            # check for duplicates

            for expr in exprs:
                valid = True
                for e_expr in expressions:
                    if e_expr == expr:
                        valid = False
                        break
                if valid:
                    expressions.append(expr)

        for expression in expressions:
            lstm_label = self.label_decoder.initial_state()
            for index in expression:
                lstm_label = lstm_label.add_input(proj_x[index])
            label_soft = self.label_w.expr(update=True) * lstm_label.output() + self.label_b.expr(update=True)
            label_index = np.argmax(label_soft.npvalue())
            labels.append(self.encodings.labels[label_index])

        return expressions, labels



import os
import unittest2
import httpretty

# Append parent dir to sys path.
parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
os.sys.path.insert(1, parentdir)

from cube.io_utils.model_store import ModelStore


XML_CONTENT_FOR_GET_LATEST_VERSION = (
    b'\xef\xbb\xbf<?xml version="2.0" encoding="utf-8"?><EnumerationResults ServiceEndpoint="'
    b'https://nlpcube.blob.core.windows.net/" ContainerName="models"><Blobs><Blob><Name>en-2.0.zip'
    b'</Name><Properties><Last-Modified>Fri, 27 Jul 2018 11:45:57 GMT</Last-Modified>'
    b'<Etag>0x8D5F3B684C5345A</Etag><Content-Length>310078470</Content-Length>'
    b'<Content-Type>application/zip</Content-Type><Content-Encoding /><Content-Language />'
    b'<Content-MD5>zrLJ7dMMbF6wYO9TaZr8GQ==</Content-MD5><Cache-Control /><Content-Disposition />'
    b'<BlobType>BlockBlob</BlobType><LeaseStatus>unlocked</LeaseStatus><LeaseState>available</LeaseState>'
    b'</Properties></Blob><Blob><Name>ro-1.0.zip</Name><Properties>'
    b'<Last-Modified>Tue, 24 Jul 2018 15:40:34 GMT</Last-Modified><Etag>0x8D5F17BCBAFCD25</Etag>'
    b'<Content-Length>310078470</Content-Length><Content-Type>application/zip</Content-Type>'
    b'<Content-Encoding /><Content-Language /><Content-MD5>zrLJ7dMMbF6wYO9TaZr8GQ==</Content-MD5>'
    b'<Cache-Control /><Content-Disposition /><BlobType>BlockBlob</BlobType>'
    b'<LeaseStatus>unlocked</LeaseStatus><LeaseState>available</LeaseState></Properties></Blob></Blobs>'
    b'<NextMarker /></EnumerationResults>'
)


class BackendTestCase(unittest2.TestCase):

    @httpretty.activate
    def test_get_latest_model_versions(self):
        httpretty.register_uri(
            httpretty.GET,
            ModelStore.MODELS_PATH_CLOUD_ALL,
            body=XML_CONTENT_FOR_GET_LATEST_VERSION
        )

        model_store = ModelStore()
        latest_model_versions = model_store.get_latest_model_versions()
        self.assertEqual(latest_model_versions, {'en': '2.0', 'ro': '1.0'})
